/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004 - 2006
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline.graphics;

// External imports
import javax.vecmath.Matrix4f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.HashSet;
import org.j3d.util.MatrixUtils;

/**
 * Common implementation of many of the capabilities required by any cull
 * stage.
 * <p>
 *
 * The class takes care of the majority of the basic implementation
 * requirements such as resizing various data structures, traversing the
 * top of the scene graph super structure (layers, viewports et al) and
 * processing offscreen sources, before sending onto the next stage.
 * All the derived class is required to do is handle traversing a single
 * scene instance and registering any pBuffer offscreens it happens to find.
 * This class will make sure that they are looked after appropriately.
 *
 * @author Justin Couch
 * @version $Revision: 3.11 $
 */
public abstract class BaseCullStage implements GraphicsCullStage
{
    /** Message when setting the active view if it contains a shared parent */
    protected static final String SHARED_VP_MSG =
        "The path from the root of the scene graph to the viewpoint contains " +
        "a SharedGroup or SharedNode instance. This is not permitted.";

    /** Message when setting the active background if it contains a shared parent */
    protected static final String SHARED_BG_MSG =
        "The path from the root of the scene graph to the background contains " +
        "a SharedGroup or SharedNode instance. This is not permitted.";

    /** Initial scene depth used for the transform stack */
    private static final int TRANSFORM_DEPTH_SIZE = 64;

    /** The increment size of the stack if it gets overflowed */
    private static final int STACK_INCREMENT = 32;

    /** Initial size of the cull list for the number of offscreen surfaces */
    protected static final int LIST_START_LENGTH = 1;

    /** The increment size of the list if it gets overflowed */
    protected static final int LIST_INCREMENT = 20;

    /** The initial size of the light list */
    protected static final int LIGHT_START_SIZE = 32;

    /** The increment size of the list if it gets overflowed */
    protected static final int LIGHT_INCREMENT = 8;

    /** If set, recurse into Shape3Ds looking for offscreen textures */
    protected boolean checkOffscreens;

    /** Buffer list of sub scenes that still need processing */
    protected PBufferTextureSource[] pendingSubscenes;

    /** Buffer list of sub scenes that still need processing */
    protected PBufferTextureSource[] pendingParents;

    /** Index of the last subscene in the list */
    protected int lastSubscene;

    /** List of the final layers that are to be passed to the output */
    protected ViewportCollection[][] outputLayers;

    /** List of the final layers that are to be passed to the output */
    protected int[] layerCounts;

    /** Index of the the next layer output */
    protected int lastOutputLayer;

    /** List of valid subscene parents as we process them */
    protected SceneGraphObject[][] validSceneParents;

    /** List of valid subscene parents as we process them */
    protected SceneGraphObject[][] outputSceneParents;

    /** List that is being used to fill values into */
    protected GraphicsCullOutputDetails[] workCullList;

    /** Index to the next output list in the validCullList */
    protected int lastOutputList;

    /** A stack used to control the depth of the transform tree */
    protected Matrix4f[] transformStack;

    /** Index to the next place to add items in the transformStack */
    protected int lastTxStack;

    /** A stack used to control the depth of the fog nodes down the tree */
    protected Fog[] fogStack;

    /** Index to the next place to add items in the fogStack */
    protected int lastFogStack;

    /** The list of lights nodes currently valid while traversing. */
    protected Light[] lightList;

    /** The list of light transforms currently valid while traversing */
    protected float[][] lightTxList;

    /** Index to the next place to add items in the lightList */
    protected int lastLight;

    /** The list of clip plane nodes currently valid while traversing. */
    protected ClipPlane[] clipList;

    /** The list of clip transforms currently valid while traversing */
    protected float[][] clipTxList;

    /** Index to the next place to add items in the clipList */
    protected int lastClip;

    /** Temp used to get texture units when looking for offscreen textures. */
    protected TextureUnit[] textureUnits;

    /** Collection of offscreen textures we've found this frame */
    protected HashSet processedPBufferTextures;

    /** Collection of textures we've found for the current parent texture */
    protected HashSet visitedNodes;

    /** Handler for the output */
    protected CulledGeometryReceiver receiver;

    /** The active parent scene of the scene we are processing. */
    protected PBufferTextureSource activeParent;

    /** Path to the current viewpoint */
    protected TransformGroup[] currentViewpointPath;

    /** Matrix used for pre-computing the view stack */
    protected Matrix4f viewMatrix1;

    /** Matrix used for pre-computing the view stack */
    protected Matrix4f viewMatrix2;

    /** Rotation matrix for the orientation provided by user */
    protected Matrix4f orientationMatrix;

    /** Storage variable for the screen orientation values */
    protected float[] screenOrientation;

    /** Storage variable for the eye offset values */
    protected float[] eyePoint;

    /** Utility for fetching RenderPass objects from their scene */
    private RenderPass[] renderPasses;

    /** Flag to say explicit screen orientation values have been provided */
    protected boolean useOrientation;

    /** Flag to say explicit eyepoint values have been provided */
    protected boolean useEyePoint;

    /** Flag indicating a shutdown of the current processing is requested */
    protected boolean terminate;

    /** Temporary variable for fetching layers from sub scenes */
    protected Layer[] layersTmp;

    /** Matrix utility code for doing inversions */
    protected MatrixUtils matrixUtils;

    /**
     * Create a basic instance of this class with the list initial internal
     * setup for the given number of renderable surfaces. The size is just an
     * initial esstimate, and is used for optimisation purposes to prevent
     * frequent array reallocations internally. As such, the number does not
     * have to be perfect, just good enough.
     *
     * @param numSurfaces Total number of surfaces to prepare rendering for
     */
    public BaseCullStage(int numSurfaces)
    {
        validSceneParents = new SceneGraphObject[numSurfaces][2];
        outputSceneParents = new SceneGraphObject[numSurfaces][2];
        pendingSubscenes = new PBufferTextureSource[numSurfaces];
        pendingParents = new PBufferTextureSource[numSurfaces];

        // Default for the most uses of a single scene/layer
        outputLayers = new ViewportCollection[numSurfaces][1];
        layerCounts = new int[1];

        for(int i = 0; i < numSurfaces; i++)
            outputLayers[i][0] = new ViewportCollection();

        layersTmp = new Layer[1];
        renderPasses = new RenderPass[6];

        textureUnits = new TextureUnit[32];
        processedPBufferTextures = new HashSet();
        visitedNodes = new HashSet();

        currentViewpointPath = new TransformGroup[TRANSFORM_DEPTH_SIZE];
        viewMatrix1 = new Matrix4f();
        viewMatrix2 = new Matrix4f();
        orientationMatrix = new Matrix4f();
        orientationMatrix.setIdentity();

        screenOrientation = new float[4];
        eyePoint = new float[3];

        useEyePoint = false;
        useOrientation = false;

        // populate the list with valid instances...

        transformStack = new Matrix4f[TRANSFORM_DEPTH_SIZE];

        // populate the list with valid instances...
        for(int i = 0; i < TRANSFORM_DEPTH_SIZE; i++)
            transformStack[i] = new Matrix4f();

        fogStack = new Fog[TRANSFORM_DEPTH_SIZE];

        lightList = new Light[LIGHT_START_SIZE];
        lightTxList = new float[LIGHT_START_SIZE][16];

        clipList = new ClipPlane[LIGHT_START_SIZE];
        clipTxList = new float[LIGHT_START_SIZE][16];

        checkOffscreens = true;

        matrixUtils = new MatrixUtils();
    }

    //---------------------------------------------------------------
    // Methods defined by CullStage
    //---------------------------------------------------------------

    /**
     * Update and cull the scenegraph defined by a set of layers. This
     * generates an ordered list of nodes to render. It will not return until
     * the culling is complete.
     *
     * @param layers The collection of layers, in order, to render
     * @param numLayers The number of valid layers to use
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    public void cull(Layer[] layers, int numLayers, int buffer)
    {
        terminate = false;

        if(outputLayers[0].length < numLayers)
        {
            outputLayers = new ViewportCollection[1][numLayers];
            for(int i = 0; i < numLayers; i++)
                outputLayers[0][i] = new ViewportCollection();

            layerCounts = new int[1];
        }

        // Pretty simple implementation - for each layer, look at each viewport
        // and then each layer within that viewport. Find the scene and
        // process it.

        int layer = 0;
        lastOutputLayer = 0;
        lastOutputList = 0;
        lastSubscene = 0;

        for(int i = 0; i < numLayers && !terminate; i++)
        {
            if(layers[i] == null)
                continue;

            switch(layers[i].getType())
            {
                case Layer.SIMPLE:
                    cullSimpleLayer((SimpleLayer)layers[i],
                                    0,
                                    layer,
                                    buffer);
                    break;

                case Layer.COMPOSITE:
                    cullCompositeLayer((CompositeLayer)layers[i],
                                       0,
                                       layer,
                                       buffer);
                    break;

                case Layer.FLAT:
                    cull2DLayer((Layer2D)layers[i], 0, layer, buffer);
                    break;

                default:
                    System.out.println("Unknown layer type to cull");
            }

            layer++;
        }

        layerCounts[0] = layer;
        lastOutputLayer++;

        if(checkOffscreens)
        {
            // May need to shift inside the loop for cases where textures
            // contain scenes which contain further textures. Play a wait and
            // see game for now.
            resizeOutputLists();

            for(int i = 0; i < lastSubscene && !terminate; i++)
            {
                activeParent = pendingSubscenes[i];
                visitedNodes.clear();

                lastOutputLayer++;

                if(processedPBufferTextures.contains(pendingSubscenes[i]))
                {
                    validSceneParents[lastOutputList][0] =
                        (SceneGraphObject)pendingSubscenes[i];
                    validSceneParents[lastOutputList][1] =
                        (SceneGraphObject)pendingParents[i];
                }
                else
                {
                    int num_layers = pendingSubscenes[i].numLayers();
                    if(layersTmp.length < num_layers)
                        layersTmp = new Layer[num_layers];

                    pendingSubscenes[i].getLayers(layersTmp);

                    // Reset these for each separate subscene
                    lastLight = 0;
                    lastClip = 0;
                    lastTxStack = 0;
                    lastFogStack = 0;

                    processSubsceneLayers(layersTmp,
                                          i + 1,
                                          num_layers,
                                          buffer);

                    processedPBufferTextures.add(pendingSubscenes[i]);

                    validSceneParents[lastOutputList][0] =
                        (SceneGraphObject)pendingSubscenes[i];
                    validSceneParents[lastOutputList][1] =
                        (SceneGraphObject)pendingParents[i];
                }

                lastOutputList++;
            }

            // Now reverse the list since we're doing a breadth-first search
            // for the subscenes, we end up with the most nested subscene last.
            // This is bad. So now reverse the order of the list so that the
            // most nested collection is first to be rendered and the main
            // screen is the last item.
            if(outputSceneParents.length < lastOutputList)
                outputSceneParents = new SceneGraphObject[lastOutputList][2];

            for(int i = 0; i < lastOutputList; i++)
            {
                outputSceneParents[i][0] =
                    validSceneParents[lastOutputList - i - 1][0];
                outputSceneParents[i][1] =
                    validSceneParents[lastOutputList - i - 1][1];
            }

        }
        else
        {
            // It is likely in this case that we're doing single-scene
            // rendering and will be doing it all the time. Possibly optimise
            // this by avoiding the copy altogether, but not 100% this won't
            // break in some corner case right now.
            outputSceneParents[0][0] = validSceneParents[0][0];
            outputSceneParents[0][1] = validSceneParents[0][1];
        }

        processedPBufferTextures.clear();

        cleanupOldRefs();

        if((receiver != null) && !terminate)
            receiver.culledOutput(outputLayers,
                                  layerCounts,
                                  lastOutputLayer,
                                  outputSceneParents,
                                  buffer);
    }

    /**
     * Register a reciever for the output of the sorter. If the value is null,
     * it will clear the currently set receiver.
     *
     * @param sgr The receiver instance to add or null
     */
    public void setCulledGeometryReceiver(CulledGeometryReceiver sgr)
    {
        receiver = sgr;
    }

    /**
     * Set the flag for whether to check for offscreen textures or not. By
     * default, this flag is set to true.
     *
     * @param state true if offscreen textures should be looked for
     */
    public void setOffscreenCheckEnabled(boolean state)
    {
        checkOffscreens = state;
    }

    /**
     * Find out what the current offscreen check state is.
     *
     * @return true if the checking is being performed
     */
    public boolean isOffscreenCheckEnabled()
    {
        return checkOffscreens;
    }

    /**
     * Set the eyepoint offset from the centre position. This is used to model
     * offset view frustums, such as multiple displays or a powerwall. This
     * method will be called with the appropriate values from the
     * RenderPipeline that this culler is inserted into.
     *
     * @param x The x axis offset
     * @param y The y axis offset
     * @param z The z axis offset
     */
    public void setEyePointOffset(float x, float y, float z)
    {
        useEyePoint = true;
        eyePoint[0] = x;
        eyePoint[1] = y;
        eyePoint[2] = z;
    }

    /**
     * Set the orientation of this screen relative to the user's normal view
     * direction. The normal orientation of the screen is along the negative
     * Z axis. This method provides and axis-angle reorientation of that
     * direction to one that is facing the screen. Typically this will just
     * involve a rotation around the Y axis of some amount (45 and 90 deg
     * being the most common used in walls and caves). This method will be
     * called with the appropriate values from the RenderPipeline that this
     * culler is inserted into.
     *
     * @param x The x axis component
     * @param y The y axis component
     * @param z The z axis component
     * @param a The angle to rotate around the axis in radians
     * @throws IllegalArgumentException The length of the axis is zero
     */
    public void setScreenOrientation(float x, float y, float z, float a)
        throws IllegalArgumentException
    {
        float n = x * x + y * y + z * z;
        if(n == 0)
            throw new IllegalArgumentException("Zero length axis vector");

        useOrientation = true;
        screenOrientation[0] = x;
        screenOrientation[1] = y;
        screenOrientation[2] = z;
        screenOrientation[3] = a;

        double rcos = Math.cos(a);
        double rsin = Math.sin(a);

        orientationMatrix.m00 = (float)(rcos + x * x * (1 - rcos));
        orientationMatrix.m01 = (float)(z * rsin + y * x * (1 - rcos));
        orientationMatrix.m02 = (float)(-y * rsin + z * x * (1 - rcos));
        orientationMatrix.m10 = (float)(-z * rsin + x * y * (1 - rcos));
        orientationMatrix.m11 = (float)(rcos + y * y * (1 - rcos));
        orientationMatrix.m12 = (float)(x * rsin + z * y * (1 - rcos));
        orientationMatrix.m20 = (float)(y * rsin + x * z * (1 - rcos));
        orientationMatrix.m21 = (float)(-x * rsin + y * z * (1 - rcos));
        orientationMatrix.m22 = (float)(rcos + z * z * (1 - rcos));
    }

    /**
     * Force a halt of the current processing. Any processing in progress
     * should exit immediately. Used to abort the current scene processing due
     * to application shutdown or complete scene replacement.
     */
    public void halt()
    {
        terminate = true;
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Update and cull the scenegraph. This generates an ordered list
     * of nodes to render. It will not return until the culling is complete.
     *
     * @param scene The scene instance to cull
     * @param subsceneId The index of the subscene in the viewport output
     * @param layerId The index of the layer in the output
     * @param viewIndex The index of the viewport in the parent
     *    ViewportCollection
     * @param layerIndex The index of the layer within the viewport
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    protected abstract void cullScene(SimpleScene scene,
                                      int subsceneId,
                                      int layerId,
                                      int viewIndex,
                                      int layerIndex,
                                      int buffer);

    /**
     * Update and cull the a single pass from a multipass rendering. This
     * generates an ordered list of nodes to render in the same was as a normal
     * scene, but with fewer items updated, such as only a single background
     * for all passes. It will not return until the culling is complete.
     *
     * @param pass The rendering pass instance to cull
     * @param subsceneId The index of the subscene in the viewport output
     * @param layerId The index of the layer in the output
     * @param viewIndex The index of the viewport in the parent
     *    ViewportCollection
     * @param layerIndex The index of the layer within the viewport
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    protected abstract void cullRenderPass(RenderPass pass,
                                           int passNumber,
                                           int subsceneId,
                                           int layerId,
                                           int viewIndex,
                                           int layerIndex,
                                           int buffer);

    /**
     * Update and cull the scenegraph of a single node without doing any other
     * processing. This generates an ordered list
     * of nodes to render. It will not return until the culling is complete.
     *
     * @param node The root node to start the cull from
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     * @return The index of the last item on the cull list
     */
    protected abstract int cullSingle(Group node, int buffer);

    /**
     * Local internal method to process a single SimpleLayer instance.
     *
     * @param layer The layer instance to process
     * @param subsceneId The index of the subscene in the viewport output
     * @param layerId The index of the layer in the output
     * @param buffer The id of the buffer to process for this round
     */
    protected void cullSimpleLayer(SimpleLayer layer,
                                   int subsceneId,
                                   int layerId,
                                   int buffer)
    {
        Viewport vp = layer.getViewport();
        outputLayers[subsceneId][layerId].numViewports = 0;
        cullViewport(vp, subsceneId, layerId, 0, buffer);
    }

    /**
     * Local internal method to process a single CompositeLayer instance.
     *
     * @param layer The layer instance to process
     * @param subsceneId The index of the subscene in the viewport output
     * @param layerId The index of the layer in the output
     * @param buffer The id of the buffer to process for this round
     */
    protected void cullCompositeLayer(CompositeLayer layer,
                                      int subsceneId,
                                      int layerId,
                                      int buffer)
    {
        int num_viewports = layer.numViewports();

        outputLayers[subsceneId][layerId].ensureCapacity(num_viewports);
        outputLayers[subsceneId][layerId].numViewports = 0;

        for(int i = 0; i < num_viewports; i++)
            cullViewport(layer.getViewport(i), subsceneId, layerId, i, buffer);
    }

    /**
     * Local internal method to process a single Layer2D instance.
     *
     * @param layer The layer instance to process
     * @param subsceneId The index of the subscene in the viewport output
     * @param layerId The index of the layer in the output
     * @param buffer The id of the buffer to process for this round
     */
    protected void cull2DLayer(Layer2D layer,
                               int subsceneId,
                               int layerId,
                               int buffer)
    {
        // Do nothing for now.
    }

    /**
     * Process a single viewport for culling.
     *
     * @param view The viewport instance to process now
     * @param subsceneId The index of the subscene in the viewport output
     * @param layerId The index of the layer in the output
     * @param viewIndex The index of the viewport in the parent
     *    ViewportCollection
     * @param buffer The id of the buffer to process for this round
     */
    protected void cullViewport(Viewport view,
                                int subsceneId,
                                int layerId,
                                int viewIndex,
                                int buffer)
    {
        if(view == null || !view.isValid())
            return;

        switch(view.getType())
        {
            case Viewport.COMPOSITE:
                outputLayers[subsceneId][layerId].numViewports++;
                cullCompositeViewport((CompositeViewport)view,
                                      subsceneId,
                                      layerId,
                                      viewIndex,
                                      buffer);
                break;

            case Viewport.SIMPLE:
                outputLayers[subsceneId][layerId].numViewports++;
                cullSimpleViewport((SimpleViewport)view,
                                   subsceneId,
                                   layerId,
                                   viewIndex,
                                   buffer);
                break;

            case Viewport.MULTIPASS:
                outputLayers[subsceneId][layerId].numViewports++;
                cullMultipassViewport((MultipassViewport)view,
                                      subsceneId,
                                      layerId,
                                      viewIndex,
                                      buffer);
                break;
        }
    }

    /**
     * Cull a composite viewport instance. A composite viewport has a number
     * of layers to work through. Each layer is held to the size of the
     * viewport.
     *
     * @param viewport The viewport instance to process
     * @param subsceneId The index of the subscene in the viewport output
     * @param layerId The index of the layer in the output
     * @param viewIndex The index of the viewport in the parent
     *    ViewportCollection
     * @param buffer The id of the buffer to process for this round
     */
    protected void cullCompositeViewport(CompositeViewport viewport,
                                         int subsceneId,
                                         int layerId,
                                         int viewIndex,
                                         int buffer)
    {
        int num = viewport.numViewportLayers();

        ViewportLayerCollection c =
            outputLayers[subsceneId][layerId].layers[viewIndex];

        c.ensureSceneCapacity(num);
        c.numBuckets = 0;
        c.numScenes = 0;
        c.numMultipass = 0;

        for(int i = 0; i < num; i++)
        {
            ViewportLayer layer = viewport.getViewportLayer(i);

            switch(layer.getType())
            {
                case ViewportLayer.SIMPLE:
                    c.useNormalScene[c.numBuckets] = true;

                    SimpleViewportLayer sl = (SimpleViewportLayer)layer;
                    cullScene(sl.getScene(),
                              subsceneId,
                              layerId,
                              viewIndex,
                              i,
                              buffer);

                    c.numScenes++;
                    c.numBuckets++;
                    break;

                case ViewportLayer.FLAT:
                    break;

                case ViewportLayer.MULTIPASS:
                    MultipassViewportLayer ml = (MultipassViewportLayer)layer;
                    cullMultipassViewportLayer(ml.getScene(),
                                               subsceneId,
                                               layerId,
                                               viewIndex,
                                               i,
                                               buffer);
                    break;
            }
        }
    }

    /**
     * Cull through a simple viewport instance. A simple viewport only has
     * a single layer, with the scene directly handled.
     *
     * @param viewport The viewport instance to process
     * @param subsceneId The index of the subscene in the viewport output
     * @param layerId The index of the layer in the output
     * @param viewIndex The index of the viewport in the parent
     *    ViewportCollection
     * @param buffer The id of the buffer to process for this round
     */
    protected void cullSimpleViewport(SimpleViewport viewport,
                                      int subsceneId,
                                      int layerId,
                                      int viewIndex,
                                      int buffer)
    {
        ViewportLayerCollection c =
            outputLayers[subsceneId][layerId].layers[viewIndex];

        c.useNormalScene[0] = true;
        c.numBuckets = 0;
        c.numScenes = 0;
        c.numMultipass = 0;

        cullScene(viewport.getScene(),
                  subsceneId,
                  layerId,
                  viewIndex,
                  0,
                  buffer);

        c.numScenes = 1;
        c.numBuckets = 1;
        c.numMultipass = 0;
        c.multipass = null;
    }

    /**
     * Cull through a simple viewport instance. A simple viewport only has
     * a single layer, with the scene directly handled.
     *
     * @param viewport The viewport instance to process
     * @param subsceneId The index of the subscene in the viewport output
     * @param layerId The index of the layer in the output
     * @param viewIndex The index of the viewport in the parent
     *    ViewportCollection
     * @param buffer The id of the buffer to process for this round
     */
    protected void cullMultipassViewport(MultipassViewport viewport,
                                         int subsceneId,
                                         int layerId,
                                         int viewIndex,
                                         int buffer)
    {
        ViewportLayerCollection c =
            outputLayers[subsceneId][layerId].layers[viewIndex];

        c.useNormalScene[0] = false;
        c.numBuckets = 0;
        c.numScenes = 0;
        c.numMultipass = 0;

        MultipassScene scene = viewport.getScene();
        int num_passes = scene.numRenderPasses();

        c.ensureMultipassCapacity(1);
        c.multipass[0].mainScene.ensureCapacity(num_passes);

        fillEnvData(scene,
                    c.multipass[c.numMultipass].mainScene.globalData,
                    buffer);

        c.multipass[c.numMultipass].mainScene.numPasses = num_passes;

        if(renderPasses.length < num_passes)
            renderPasses = new RenderPass[num_passes];

        scene.getRenderPasses(renderPasses);

        for(int i = 0; i < num_passes; i++)
        {
            cullRenderPass(renderPasses[i],
                           i,
                           subsceneId,
                           layerId,
                           viewIndex,
                           0,
                           buffer);
        }

        c.numBuckets = 1;
        c.numScenes = 0;
        c.numMultipass = 1;
    }

    /**
     * Cull through a simple viewport instance. A simple viewport only has
     * a single layer, with the scene directly handled.
     *
     * @param scene The scene instance to process
     * @param subsceneId The index of the subscene in the viewport output
     * @param layerId The index of the layer in the output
     * @param viewIndex The index of the viewport in the parent
     *    ViewportCollection
     * @param layerIndex The index of the layer within the viewport
     * @param buffer The id of the buffer to process for this round
     */
    protected void cullMultipassViewportLayer(MultipassScene scene,
                                             int subsceneId,
                                             int layerId,
                                             int viewIndex,
                                             int layerIndex,
                                             int buffer)
    {
        ViewportLayerCollection c =
            outputLayers[subsceneId][layerId].layers[viewIndex];

        c.useNormalScene[c.numBuckets] = false;

        int num_passes = scene.numRenderPasses();

        c.ensureMultipassCapacity(1);
        c.multipass[0].mainScene.ensureCapacity(num_passes);

        fillEnvData(scene,
                    c.multipass[c.numMultipass].mainScene.globalData,
                    buffer);

        c.multipass[c.numMultipass].mainScene.numPasses = num_passes;

        if(renderPasses.length < num_passes)
            renderPasses = new RenderPass[num_passes];

        scene.getRenderPasses(renderPasses);

        for(int i = 0; i < num_passes; i++)
        {
            cullRenderPass(renderPasses[i],
                           i,
                           subsceneId,
                           layerId,
                           viewIndex,
                           layerIndex,
                           buffer);
        }

        c.numBuckets++;
        c.numMultipass++;
    }

    /**
     * Process the layers of a pbuffer texture source.
     *
     * @param subsceneId The index of the subscene in the viewport output
     * @param layerId The index of the layer in the output
     */
    protected void processSubsceneLayers(Layer[] layers,
                                         int subsceneId,
                                         int numLayers,
                                         int buffer)
    {
        for(int i = 0; i < numLayers; i++)
        {
            if(layers[i] == null)
                continue;

            switch(layers[i].getType())
            {
                case Layer.SIMPLE:
                    cullSimpleLayer((SimpleLayer)layers[i],
                                    subsceneId,
                                    i,
                                    buffer);
                    break;

                case Layer.COMPOSITE:
                    cullCompositeLayer((CompositeLayer)layers[i],
                                       subsceneId,
                                       i,
                                       buffer);
                    break;

                case Layer.FLAT:
                    cull2DLayer((Layer2D)layers[i],
                                subsceneId,
                                i,
                                buffer);
                    break;

                default:
                    System.out.println("Unknown layer type to cull");
            }
        }
    }

    /**
     * Take a simple scene and fill in a GraphicsEnvironmentData instance.
     *
     * @param scene The scene to take data from
     * @param envData Data instance to copy it to
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    protected void fillEnvData(SimpleScene scene,
                               GraphicsEnvironmentData envData,
                               int buffer)
    {
        ViewEnvironment view = scene.getViewEnvironment();

        envData.userData = scene.getUserData();
        envData.effectsProcessor = scene.getRenderEffectsProcessor();
        envData.viewpoint = scene.getActiveView();
        envData.background = scene.getActiveBackground();

        envData.viewProjectionType = view.getProjectionType();
        envData.useStereo = view.getStereoEnabled();

        Fog fog = scene.getActiveFog();
        if((fog != null) && fog.isEnabled())
            envData.fog = fog;
        else
            envData.fog = null;

        envData.eyeOffset[0] = eyePoint[0];
        envData.eyeOffset[1] = eyePoint[1];
        envData.eyeOffset[2] = eyePoint[2];

        // Calculate frustum parameters based on view environment
        int[] viewport = view.getViewportDimensions();
        envData.setViewport(viewport);

        view.generateViewFrustum(envData.viewFrustum);

        if(useEyePoint)
        {
            envData.viewFrustum[0] -= eyePoint[0];
            envData.viewFrustum[1] -= eyePoint[0];
            envData.viewFrustum[2] -= eyePoint[1];
            envData.viewFrustum[3] -= eyePoint[1];
            envData.viewFrustum[4] -= eyePoint[2];
            envData.viewFrustum[5] -= eyePoint[2];
        }

        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        Node parent = envData.viewpoint.getParent();
        int path_depth = 0;

        while(parent != null)
        {
            if((parent instanceof SharedGroup) ||
               (parent instanceof SharedNode))
                throw new IllegalArgumentException(SHARED_VP_MSG);

            parent = parent.getParent();
            path_depth++;
        }

        if(currentViewpointPath.length < path_depth)
            currentViewpointPath = new TransformGroup[path_depth];

        // start at the VP and work our way upwards.
        parent = envData.viewpoint.getParent();
        int last_tx = 0;
        while(parent != null)
        {
            if(parent instanceof TransformGroup)
                currentViewpointPath[last_tx++] = (TransformGroup)parent;

            parent = parent.getParent();
        }

        viewMatrix2.setIdentity();

        if(last_tx != 0)
        {
            for(int i = last_tx - 1; i >= 0; i--)
            {
                currentViewpointPath[i].getTransform(viewMatrix1);
                viewMatrix2.mul(viewMatrix1);
            }
        }

        if(useOrientation)
            viewMatrix2.mul(orientationMatrix);

        envData.viewTransform.set(viewMatrix2);

        matrixUtils.inverse(envData.viewTransform, viewMatrix2);
        envData.cameraTransform[0] = viewMatrix2.m00;
        envData.cameraTransform[1] = viewMatrix2.m10;
        envData.cameraTransform[2] = viewMatrix2.m20;
        envData.cameraTransform[3] = viewMatrix2.m30;
        envData.cameraTransform[4] = viewMatrix2.m01;
        envData.cameraTransform[5] = viewMatrix2.m11;
        envData.cameraTransform[6] = viewMatrix2.m21;
        envData.cameraTransform[7] = viewMatrix2.m31;
        envData.cameraTransform[8] = viewMatrix2.m02;
        envData.cameraTransform[9] = viewMatrix2.m12;
        envData.cameraTransform[10] = viewMatrix2.m22;
        envData.cameraTransform[11] = viewMatrix2.m32;
        envData.cameraTransform[12] = viewMatrix2.m03;
        envData.cameraTransform[13] = viewMatrix2.m13;
        envData.cameraTransform[14] = viewMatrix2.m23;
        envData.cameraTransform[15] = viewMatrix2.m33;

        // Now do almost the same thing with the background node. Only
        // difference is that this time we deliberately set the translation
        // components to zero.
        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        if(envData.background != null)
        {
            parent = envData.background.getParent();
            path_depth = 0;

            while(parent != null)
            {
                if((parent instanceof SharedGroup) ||
                   (parent instanceof SharedNode))
                    throw new IllegalArgumentException(SHARED_BG_MSG);

                parent = parent.getParent();
                path_depth++;
            }

            // reuse the VP path for this one
            if(currentViewpointPath.length < path_depth)
                currentViewpointPath = new TransformGroup[path_depth];

            // start at the background and work our way upwards.
            parent = envData.background.getParent();
            last_tx = 0;
            while(parent != null)
            {
                if(parent instanceof TransformGroup)
                    currentViewpointPath[last_tx++] = (TransformGroup)parent;

                parent = parent.getParent();
            }

            viewMatrix2.setIdentity();

            if(last_tx != 0)
            {
                for(int i = last_tx - 1; i >= 0; i--)
                {
                    currentViewpointPath[i].getTransform(viewMatrix1);
                    viewMatrix2.mul(viewMatrix1);
                }
            }

            viewMatrix2.m30 = 0;
            viewMatrix2.m31 = 0;
            viewMatrix2.m32 = 0;

            viewMatrix2.mul(envData.viewTransform);

            envData.backgroundTransform[0] = viewMatrix2.m00;
            envData.backgroundTransform[1] = viewMatrix2.m01;
            envData.backgroundTransform[2] = viewMatrix2.m02;
            envData.backgroundTransform[3] = 0;
            envData.backgroundTransform[4] = viewMatrix2.m10;
            envData.backgroundTransform[5] = viewMatrix2.m11;
            envData.backgroundTransform[6] = viewMatrix2.m12;
            envData.backgroundTransform[7] = 0;
            envData.backgroundTransform[8] = viewMatrix2.m20;
            envData.backgroundTransform[9] = viewMatrix2.m21;
            envData.backgroundTransform[10] = viewMatrix2.m22;
            envData.backgroundTransform[11] = 0;
            envData.backgroundTransform[12] = viewMatrix2.m30;
            envData.backgroundTransform[13] = viewMatrix2.m31;
            envData.backgroundTransform[14] = viewMatrix2.m32;
            envData.backgroundTransform[15] = viewMatrix2.m33;

            setupBackgroundFrustum(0.01,
                                   1.1,
                                   view.getFieldOfView(),
                                   envData,
                                   buffer);
        }
    }

    /**
     * Take a simple scene and fill in a GraphicsEnvironmentData instance.
     *
     * @param scene The scene to take data from
     * @param envData Data instance to copy it to
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    protected void fillEnvData(MultipassScene scene,
                               GraphicsEnvironmentData envData,
                               int buffer)
    {
        ViewEnvironment view = scene.getViewEnvironment();

        envData.userData = scene.getUserData();
        envData.effectsProcessor = scene.getRenderEffectsProcessor();
        envData.viewpoint = null;
        envData.background = scene.getActiveBackground();

        envData.viewProjectionType = view.getProjectionType();
        envData.useStereo = view.getStereoEnabled();
        envData.fog = null;

        envData.eyeOffset[0] = eyePoint[0];
        envData.eyeOffset[1] = eyePoint[1];
        envData.eyeOffset[2] = eyePoint[2];

        // Calculate frustum parameters based on view environment
        int[] viewport = view.getViewportDimensions();
        envData.setViewport(viewport);

        view.generateViewFrustum(envData.viewFrustum);

        if(useEyePoint)
        {
            envData.viewFrustum[0] -= eyePoint[0];
            envData.viewFrustum[1] -= eyePoint[0];
            envData.viewFrustum[2] -= eyePoint[1];
            envData.viewFrustum[3] -= eyePoint[1];
            envData.viewFrustum[4] -= eyePoint[2];
            envData.viewFrustum[5] -= eyePoint[2];
        }
    }

    /**
     * Take a simple scene and fill in a GraphicsEnvironmentData instance.
     *
     * @param scene The scene to take data from
     * @param envData Data instance to copy it to
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    protected void fillEnvData(RenderPass pass,
                               GraphicsEnvironmentData envData,
                               BufferDetails bufferData,
                               int buffer)
    {
        bufferData.colorBufferState = pass.getColorBufferState();
        bufferData.depthBufferState = pass.getDepthBufferState();
        bufferData.stencilBufferState = pass.getStencilBufferState();
        bufferData.accumBufferState = pass.getAccumulationBufferState();

        ViewEnvironment view = pass.getViewEnvironment();

        envData.userData = pass.getUserData();
        envData.effectsProcessor = null;
        envData.viewpoint = pass.getActiveView();
        envData.background = null;

        envData.viewProjectionType = view.getProjectionType();
        envData.useStereo = false;

        // Not sure how valid an assumption this is when dealing
        // with the individual rendering passes.
        envData.eyeOffset[0] = eyePoint[0];
        envData.eyeOffset[1] = eyePoint[1];
        envData.eyeOffset[2] = eyePoint[2];

        // Calculate frustum parameters based on view environment
        int[] viewport = view.getViewportDimensions();
        envData.setViewport(viewport);

        view.generateViewFrustum(envData.viewFrustum);

        if(useEyePoint)
        {
            envData.viewFrustum[0] -= eyePoint[0];
            envData.viewFrustum[1] -= eyePoint[0];
            envData.viewFrustum[2] -= eyePoint[1];
            envData.viewFrustum[3] -= eyePoint[1];
            envData.viewFrustum[4] -= eyePoint[2];
            envData.viewFrustum[5] -= eyePoint[2];
        }

        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        Node parent = envData.viewpoint.getParent();
        int path_depth = 0;

        while(parent != null)
        {
            if((parent instanceof SharedGroup) ||
               (parent instanceof SharedNode))
                throw new IllegalArgumentException(SHARED_VP_MSG);

            parent = parent.getParent();
            path_depth++;
        }

        if(currentViewpointPath.length < path_depth)
            currentViewpointPath = new TransformGroup[path_depth];

        // start at the VP and work our way upwards.
        parent = envData.viewpoint.getParent();
        int last_tx = 0;
        while(parent != null)
        {
            if(parent instanceof TransformGroup)
                currentViewpointPath[last_tx++] = (TransformGroup)parent;

            parent = parent.getParent();
        }

        viewMatrix2.setIdentity();

        if(last_tx != 0)
        {
            for(int i = last_tx - 1; i >= 0; i--)
            {
                currentViewpointPath[i].getTransform(viewMatrix1);
                viewMatrix2.mul(viewMatrix1);
            }
        }

        if(useOrientation)
            viewMatrix2.mul(orientationMatrix);

        envData.viewTransform.set(viewMatrix2);

        matrixUtils.inverse(envData.viewTransform, viewMatrix2);
        envData.cameraTransform[0] = viewMatrix2.m00;
        envData.cameraTransform[1] = viewMatrix2.m10;
        envData.cameraTransform[2] = viewMatrix2.m20;
        envData.cameraTransform[3] = viewMatrix2.m30;
        envData.cameraTransform[4] = viewMatrix2.m01;
        envData.cameraTransform[5] = viewMatrix2.m11;
        envData.cameraTransform[6] = viewMatrix2.m21;
        envData.cameraTransform[7] = viewMatrix2.m31;
        envData.cameraTransform[8] = viewMatrix2.m02;
        envData.cameraTransform[9] = viewMatrix2.m12;
        envData.cameraTransform[10] = viewMatrix2.m22;
        envData.cameraTransform[11] = viewMatrix2.m32;
        envData.cameraTransform[12] = viewMatrix2.m03;
        envData.cameraTransform[13] = viewMatrix2.m13;
        envData.cameraTransform[14] = viewMatrix2.m23;
        envData.cameraTransform[15] = viewMatrix2.m33;
    }

    /**
     * Check a shape node for the offscreen textures that may be present. If
     * some are found, queue them up for processing.
     *
     * @param shape The object instance to process
     * @param buffer The id of the buffer to process for this round
     */
    protected void checkForOffscreens(Shape3D shape, int buffer)
    {
        Appearance app = shape.getAppearance();

        if(app == null)
            return;

        int num_units = app.numTextureUnits();
        if(num_units == 0)
            return;

        app.getTextureUnits(textureUnits);

        for(int i =0; i < num_units; i++)
        {
            Texture tex = textureUnits[i].getTexture();

            // Only queue up the texture for repainting if it says it needs a
            // repaint and that we haven't already found this instance in this
            // frame.
            if(tex instanceof PBufferTextureSource)
            {
                if(!visitedNodes.contains(tex) &&
                   ((PBufferTextureSource)tex).isRepaintRequired())
                {
                    resizeOffscreenList();
                    pendingSubscenes[lastSubscene] = (PBufferTextureSource)tex;
                    pendingParents[lastSubscene] = activeParent;
                    visitedNodes.add(tex);
                    lastSubscene++;
                }
            }
            else if(tex instanceof MultipassTextureSource)
            {
                // Fill in multipass offscreen texture
System.out.println("Null cull stage does not process multipass textures yet");
            }
        }
    }

    /**
     * Clean up the unused resources after the end of the cull process. This
     * releases any references that are no longer needed, and may have been
     * kept from the previous culling pass. This method is called just before
     * the output is sent on to the sorting stage.
     * <p>
     *
     * This base implementation cleans up the outputLayers list. Derived
     * classes that override this method for their own needs should also
     * make sure to call this method too.
     */
    protected void cleanupOldRefs()
    {
        for(int i = 0; i < outputLayers.length; i++)
        {
            for(int j = 0; j < outputLayers[i].length; j++)
            {
                if(outputLayers[i][j] == null)
                    break;

                for(int k = 0; k < outputLayers[i][j].numViewports; k++)
                {
                    ViewportLayerCollection c = outputLayers[i][j].layers[k];

                    for(int l = 0; l < c.numScenes; l++)
                    {
                        SceneRenderBucket b = c.scenes[l];

                        for(int m = b.numNodes; m <  b.nodes.length; m++)
                        {
                            if(b.nodes[m] == null)
                                break;

                            b.nodes[m].renderable = null;
                            b.nodes[m].localFog   = null;
                            b.nodes[m].customData = null;

                            // Another option we could use here is to clear
                            // the renference to the details object too. May
                            // want to have a system property for that.
                            for(int n = 0; n < b.nodes[m].numLights; n++)
                                b.nodes[m].lights[n].clear();


                            for(int n = 0; n < b.nodes[m].numClipPlanes; n++)
                                b.nodes[m].clipPlanes[n].clear();

                            b.nodes[m].numLights = 0;
                            b.nodes[m].numClipPlanes = 0;
                        }
                    }
                }
            }
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    protected final void resizeCullList(int cur_size)
    {
        if((cur_size + 1) == workCullList.length)
        {
            int old_size = workCullList.length;
            int new_size = old_size + LIST_INCREMENT;

            GraphicsCullOutputDetails[] tmp_nodes = new GraphicsCullOutputDetails[new_size];

            System.arraycopy(workCullList, 0, tmp_nodes, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_nodes[i] = new GraphicsCullOutputDetails();

            workCullList = tmp_nodes;
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    protected final void resizeOutputLists()
    {
        if((lastSubscene + 1) < validSceneParents.length)
            return;

        int old_size = validSceneParents.length;
        int new_size = old_size + LIST_INCREMENT;

        SceneGraphObject[][] tmp_obj = new SceneGraphObject[new_size][];

        System.arraycopy(validSceneParents, 0, tmp_obj, 0, old_size);

        for(int i = old_size; i < new_size; i++)
            tmp_obj[i] = new SceneGraphObject[2];

        validSceneParents = tmp_obj;

        ViewportCollection[][] tmp_vc = new ViewportCollection[new_size][];
        int[] tmp_lc = new int[new_size];

        System.arraycopy(outputLayers, 0, tmp_vc, 0, outputLayers.length);
        System.arraycopy(layerCounts, 0, tmp_lc, 0, outputLayers.length);

        for(int i = outputLayers.length; i < new_size; i++)
            tmp_vc[i] = new ViewportCollection[1];

        outputLayers = tmp_vc;
        layerCounts = tmp_lc;
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    protected final void resizeOffscreenList()
    {
        if((lastSubscene + 1) == pendingSubscenes.length)
        {
            int old_size = pendingSubscenes.length;
            int new_size = old_size + LIST_INCREMENT;

            PBufferTextureSource[] tmp_nodes =
                new PBufferTextureSource[new_size];

            System.arraycopy(pendingSubscenes, 0, tmp_nodes, 0, old_size);

            pendingSubscenes = tmp_nodes;

            tmp_nodes = new PBufferTextureSource[new_size];

            System.arraycopy(pendingParents, 0, tmp_nodes, 0, old_size);

            pendingParents = tmp_nodes;
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    protected final void resizeLightList()
    {
        if((lastLight + 1) == lightList.length)
        {
            int old_size = lightList.length;
            int new_size = old_size + LIST_INCREMENT;

            Light[] tmp_nodes = new Light[new_size];

            System.arraycopy(lightList, 0, tmp_nodes, 0, old_size);

            lightList = tmp_nodes;

            float[][] tmp_tx  = new float[new_size][];
            System.arraycopy(lightTxList, 0, tmp_tx, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_tx[i] = new float[16];

            lightTxList = tmp_tx;
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    protected final void resizeClipList()
    {
        if((lastClip + 1) == clipList.length)
        {
            int old_size = clipList.length;
            int new_size = old_size + LIST_INCREMENT;

            ClipPlane[] tmp_nodes = new ClipPlane[new_size];

            System.arraycopy(clipList, 0, tmp_nodes, 0, old_size);

            clipList = tmp_nodes;

            float[][] tmp_tx  = new float[new_size][];
            System.arraycopy(clipTxList, 0, tmp_tx, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_tx[i] = new float[16];

            clipTxList = tmp_tx;
        }
    }

    /**
     * Resize the transform stack if needed. Marked as final in order to
     * encourage the compiler to inline the code for faster execution
     */
    protected final void resizeStack()
    {
        if((lastTxStack + 1) == transformStack.length)
        {
            int old_size = transformStack.length;
            int new_size = old_size + STACK_INCREMENT;

            Matrix4f[] tmp_matrix = new Matrix4f[new_size];

            System.arraycopy(transformStack, 0, tmp_matrix, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_matrix[i] = new Matrix4f();

            transformStack = tmp_matrix;
        }
    }

    /**
     * Resize the fog stack if needed. Marked as final in order to
     * encourage the compiler to inline the code for faster execution
     */
    protected final void resizeFogStack()
    {
        if((lastFogStack + 1) == fogStack.length)
        {
            int old_size = fogStack.length;
            int new_size = old_size + STACK_INCREMENT;

            Fog[] tmp_fog = new Fog[new_size];

            System.arraycopy(fogStack, 0, tmp_fog, 0, old_size);

            fogStack = tmp_fog;
        }
    }

    /**
     * From the current view setup of FoV, near and far clipping distances and
     * the aspectRatio ratio, generate the 6 parameters that describe a view
     * frustum. These parameters are what could be used as arguments to the
     * glFrustum call. The parameter order is:
     * [x min, x max, y min, y max, z near, z far]
     *
     * @param nearClip The distance to the near clip plane
     * @param farClip The distance to the far clip plane
     * @param fov The field of view to use
     * @param data The environment data instance to fill in
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    private void setupBackgroundFrustum(double nearClip,
                                        double farClip,
                                        double fov,
                                        GraphicsEnvironmentData data,
                                        int buffer)
    {
        double xmin = 0;
        double xmax = 0;
        double ymin = 0;
        double ymax = 0;

        double aspect_ratio = (double) data.viewport[GraphicsEnvironmentData.VIEW_WIDTH] /
                              data.viewport[GraphicsEnvironmentData.VIEW_HEIGHT];

        ymax = nearClip * Math.tan(fov * Math.PI / 360.0);
        ymin = -ymax;

        xmin = ymin * aspect_ratio;
        xmax = ymax * aspect_ratio;

        data.backgroundFrustum[0] = xmin;
        data.backgroundFrustum[1] = xmax;
        data.backgroundFrustum[2] = ymin;
        data.backgroundFrustum[3] = ymax;
        data.backgroundFrustum[4] = nearClip;
        data.backgroundFrustum[5] = farClip;
    }
}
