/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline.graphics;

// External imports
import java.util.Arrays;
import java.util.HashMap;

import javax.vecmath.Matrix4f;

import javax.media.opengl.GL;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.HashSet;

import org.j3d.aviatrix3d.pipeline.RenderOp;

/**
 * Implementation of common functionality used by sort stage that implement
 * state sorting.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 3.4 $
 */
public abstract class BaseStateSortStage extends BaseSortStage
{
    /** The guess at the number of components in a shape */
    protected static final int GUESS_NUM_COMPONENTS = 5;

    /** Listing of int to Integer equivalents for texture IDs */
    protected static final Integer[] TEX_IDS;

    /** Comparator instance for the state sorting */
    protected StateSortComparator stateComparator;

    // Listing of the current items, while we're setting up the state handling.

    /** Temp array for holdin new lights on the object being processed */
    protected VisualDetails[] newLights;

    /** Temp array for fetching lights objects from the HashSets below. */
    protected Object[] lightTmp;

    /** Set of current lights that are being processed */
    protected HashSet currentLights;

    /** Set for holding the different between the current and new sets */
    protected HashSet oldLights;

    /** Set to hold onto lights that are common between last node and this */
    protected HashSet keepLights;

    /** Mapping of the VisualDetails instance to the global ID */
    protected HashMap lightIdMap;

    /** Temp array for holding new clip planes on the object being processed */
    protected VisualDetails[] newClipPlanes;

    /** Temp array for fetching clipPlane objects from the HashSets below. */
    protected Object[] clipTmp;

    /** Set of current clip planes that are being processed */
    protected HashSet currentClipPlanes;

    /** Set for holding the difference between the current and new sets */
    protected HashSet oldClipPlanes;

    /** Set to hold onto clip planes that are common between last node and this */
    protected HashSet keepClipPlanes;

    /** Mapping of the VisualDetails instance to the global ID */
    private HashMap clipIdMap;

    /** The currently valid appearance */
    protected Appearance currentAppearance;

    /** The currently valid material */
    protected Material currentMaterial;

    /** The currently valid polygon attributes */
    protected PolygonAttributes currentPolyAttr;

    /** The currently valid line attributes */
    protected LineAttributes currentLineAttr;

    /** The currently valid point attributes */
    protected PointAttributes currentPointAttr;

    /** The currently valid blend attributes */
    protected BlendAttributes currentBlendAttr;

    /** The currently valid depth buffer attributes */
    protected DepthAttributes currentDepthAttr;

    /** The currently valid stencil buffer attributes */
    protected StencilAttributes currentStencilAttr;

    /** The currently valid GLSlang shader*/
    protected Shader currentShader;

    /** The currently valid shader program (GLSLang)*/
    protected ShaderProgram currentShaderProgram;

    /** The currently valid shader arguments (GLSLang)*/
    protected ShaderArguments currentShaderArgs;

    /** The currently valid vertex shader */
    protected VertexShader currentVtxShader;

    /** The currently valid fragment shader */
    protected FragmentShader currentFragShader;

    /** The currently valid local fog */
    protected Fog currentFog;

    /** temp array for fetching textures from the appearance */
    protected TextureUnit[] textureTmp;

    /** Listing of currently valid textures */
    protected TextureUnit[] currentTextures;

    /** Number of valid textures currently active */
    protected int numTextures;

    /**
     * Static constructor to initialise some of the constants
     */
    static
    {
        TEX_IDS = new Integer[32];
        for(int i = 0; i < 32; i++)
            TEX_IDS[i] = new Integer(GL.GL_TEXTURE0 + i);
    }

    /**
     * Create an empty sorting stage that initialises the internal structures
     * to assume that there is a minumum number of surfaces, both on and
     * offscreen.
     *
     */
    protected BaseStateSortStage(int numSurfaces)
    {
        super(numSurfaces);

        stateComparator = new StateSortComparator();

        currentLights = new HashSet();
        keepLights = new HashSet();
        oldLights = new HashSet();
        lightIdMap = new HashMap();
        lightTmp = new Object[24];
        newLights = new VisualDetails[32];

        currentClipPlanes = new HashSet();
        keepClipPlanes = new HashSet();
        oldClipPlanes = new HashSet();
        clipIdMap = new HashMap();
        clipTmp = new Object[24];
        newClipPlanes = new VisualDetails[32];

        // 32 is max textures that OpenGL can support.
        currentTextures = new TextureUnit[32];
        textureTmp = new TextureUnit[32];
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * Append a single object onto the existing list. Checks versus existing
     * state and starts and stops as required.
     *
     * @param node The node instance to look at appending
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int appendObject(GraphicsCullOutputDetails node,
                               GraphicsInstructions instr,
                               int offset)
    {
        int idx = offset;

        idx = appendLights(node, instr, idx);
        idx = appendClipPlanes(node, instr, idx);
        idx = updateFog(node.localFog, instr, idx);

        // Finally, map the shape instances
        if(node.renderable instanceof Shape3D)
        {
            Shape3D shape = (Shape3D)node.renderable;
            Appearance app = shape.getAppearance();
            Geometry geom = shape.getGeometry();

            if(currentAppearance != app)
            {
                // If the new state is now null, but before it wasn't then
                // we need to clear all the current visual state. Otherwise
                // we have to change the state from the existing one and
                // setup the new state.
                if(app == null)
                {
                    idx = cleanupVisuals(instr, idx);
                }
                else
                {
                    idx = updateMaterial(app, instr, idx);
                    idx = updatePolyAttribs(app, instr, idx);
                    idx = updateLineAttribs(app, instr, idx);
                    idx = updatePointAttribs(app, instr, idx);
                    idx = updateBlendAttribs(app, instr, idx);
                    idx = updateDepthAttribs(app, instr, idx);
                    idx = updateStencilAttribs(app, instr, idx);
                    idx = updateTextures(app, instr, idx);
                    idx = updateShader(app, instr, idx);
                }

                currentAppearance = app;
            }

            instr.renderList[idx].renderable = geom;

            Matrix4f tx = node.transform;

            instr.renderList[idx].transform[0] = tx.m00;
            instr.renderList[idx].transform[1] = tx.m10;
            instr.renderList[idx].transform[2] = tx.m20;
            instr.renderList[idx].transform[3] = tx.m30;

            instr.renderList[idx].transform[4] = tx.m01;
            instr.renderList[idx].transform[5] = tx.m11;
            instr.renderList[idx].transform[6] = tx.m21;
            instr.renderList[idx].transform[7] = tx.m31;

            instr.renderList[idx].transform[8] = tx.m02;
            instr.renderList[idx].transform[9] = tx.m12;
            instr.renderList[idx].transform[10] = tx.m22;
            instr.renderList[idx].transform[11] = tx.m32;

            instr.renderList[idx].transform[12] = tx.m03;
            instr.renderList[idx].transform[13] = tx.m13;
            instr.renderList[idx].transform[14] = tx.m23;
            instr.renderList[idx].transform[15] = tx.m33;

            instr.renderOps[idx] = RenderOp.RENDER_GEOMETRY;
            idx++;
        }
        else if(node.renderable instanceof CustomRenderable)
        {
            // forcefully remove anything left
            idx = cleanupVisuals(instr, idx);

            instr.renderList[idx].renderable = node.renderable;
            instr.renderList[idx].instructions = node.customData;

            Matrix4f tx = node.transform;

            instr.renderList[idx].transform[0] = tx.m00;
            instr.renderList[idx].transform[1] = tx.m10;
            instr.renderList[idx].transform[2] = tx.m20;
            instr.renderList[idx].transform[3] = tx.m30;

            instr.renderList[idx].transform[4] = tx.m01;
            instr.renderList[idx].transform[5] = tx.m11;
            instr.renderList[idx].transform[6] = tx.m21;
            instr.renderList[idx].transform[7] = tx.m31;

            instr.renderList[idx].transform[8] = tx.m02;
            instr.renderList[idx].transform[9] = tx.m12;
            instr.renderList[idx].transform[10] = tx.m22;
            instr.renderList[idx].transform[11] = tx.m32;

            instr.renderList[idx].transform[12] = tx.m03;
            instr.renderList[idx].transform[13] = tx.m13;
            instr.renderList[idx].transform[14] = tx.m23;
            instr.renderList[idx].transform[15] = tx.m33;

            instr.renderOps[idx] = RenderOp.RENDER_CUSTOM;
            idx++;
        }
        else
        {
            System.out.println("Non-shape node in sorter " + node.renderable);
        }

        return idx;
    }

    /**
     * Append the lights onto the existing list. Checks versus existing
     * state and starts and stops as required.
     *
     * @param node The node instance to look at appending
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int appendLights(GraphicsCullOutputDetails node,
                               GraphicsInstructions instr,
                               int offset)
    {
        if(node.numLights == 0 && currentLights.size() == 0)
            return offset;

        int idx = offset;

        // First drop all the lights into the queue and turn the state off on
        // the no-longer used ones.
        int new_count = 0;

        for(int i = 0; i < node.numLights; i++)
        {
            VisualDetails ld = node.lights[i];

            if(currentLights.contains(ld))
                keepLights.add(ld);
            else
                newLights[new_count++] = ld;
        }

        currentLights.retainAll(keepLights, oldLights);
        keepLights.clear();

        // Unmap the old lights.
        int num_old = oldLights.size();
        if(num_old != 0)
        {
            if(num_old > lightTmp.length)
                lightTmp = new Object[num_old];

            oldLights.toArray(lightTmp);
            oldLights.clear();

            for(int i = 0; i < num_old; i++)
            {
                VisualDetails ld = (VisualDetails)lightTmp[i];
                Integer l_id = (Integer)lightIdMap.remove(ld);

                instr.renderList[idx].renderable = ld.getRenderable();
                instr.renderList[idx].id = l_id.intValue();

                // Don't need the transform for the light stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_LIGHT;
                idx++;
                lightTmp[i] = null;
            }
        }

        // Map the new lights.
        if(new_count != 0)
        {
            for(int i = 0; i < new_count; i++)
            {
                VisualDetails ld = newLights[i];
                currentLights.add(ld);

                instr.renderList[idx].renderable = ld.getRenderable();

                System.arraycopy(ld.getTransform(),
                                 0,
                                 instr.renderList[idx].transform,
                                 0,
                                 16);

                lightIdMap.put(ld, new Integer(lastGlobalId));
                instr.renderList[idx].id = lastGlobalId++;
                instr.renderOps[idx] = RenderOp.START_LIGHT;
                idx++;
                newLights[i] = null;
            }
        }

        return idx;
    }

    /**
     * Append the clip planess onto the existing list. Checks versus existing
     * state and starts and stops as required.
     *
     * @param node The node instance to look at appending
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int appendClipPlanes(GraphicsCullOutputDetails node,
                                   GraphicsInstructions instr,
                                   int offset)
    {
        if(node.numClipPlanes == 0 && currentClipPlanes.size() == 0)
            return offset;

        int idx = offset;

        // Drop all the clips into the queue and turn the state off on
        // the no-longer used ones.
        int new_count = 0;

        for(int i = 0; i < node.numClipPlanes; i++)
        {
            VisualDetails ld = node.clipPlanes[i];

            if(currentClipPlanes.contains(ld))
                keepClipPlanes.add(ld);
            else
                newClipPlanes[new_count++] = ld;
        }

        currentClipPlanes.retainAll(keepClipPlanes, oldClipPlanes);
        keepClipPlanes.clear();

        // Unmap the old clips.
        int num_old = oldClipPlanes.size();
        if(num_old != 0)
        {
            if(num_old > clipTmp.length)
                clipTmp = new Object[num_old];

            oldClipPlanes.toArray(clipTmp);
            oldClipPlanes.clear();

            for(int i = 0; i < num_old; i++)
            {
                VisualDetails ld = (VisualDetails)clipTmp[i];
                Integer l_id = (Integer)clipIdMap.remove(ld);

                instr.renderList[idx].renderable = ld.getRenderable();
                instr.renderList[idx].id = l_id.intValue();

                // Don't need the transform for the clip stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_CLIP_PLANE;
                idx++;
                clipTmp[i] = null;
            }
        }

        // Map the new clips.
        if(new_count != 0)
        {
            for(int i = 0; i < new_count; i++)
            {
                VisualDetails ld = newClipPlanes[i];
                currentClipPlanes.add(ld);

                instr.renderList[idx].renderable = ld.getRenderable();

                System.arraycopy(ld.getTransform(),
                                 0,
                                 instr.renderList[idx].transform,
                                 0,
                                 16);

                clipIdMap.put(ld, new Integer(lastGlobalId));
                instr.renderList[idx].id = lastGlobalId++;
                instr.renderOps[idx] = RenderOp.START_CLIP_PLANE;
                idx++;
                newClipPlanes[i] = null;
            }
        }

        return idx;
    }

    /**
     * Check on the material setup and change as necessary.
     *
     * @param fog The Fog instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int updateFog(Fog fog, GraphicsInstructions instr, int offset)
    {
        int idx = offset;

        if(fog != currentFog)
        {
            if(fog != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!fog.equals(currentFog))
                {
                    // Clear the old state if it is set
                    if(currentFog != null)
                    {
                        instr.renderList[idx].renderable = currentFog;
                        instr.renderOps[idx] = RenderOp.STOP_FOG;
                        idx++;
                    }

                    instr.renderList[idx].renderable = fog;
                    instr.renderOps[idx] = RenderOp.START_FOG;
                    idx++;
                }

                currentFog = fog;
            }
            else
            {
                // The new fogerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentFog;
                instr.renderOps[idx] = RenderOp.STOP_FOG;
                idx++;

                currentFog = null;
            }
        }

        return idx;
    }


    /**
     * Check on the shader setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int updateShader(Appearance app,
                               GraphicsInstructions instr,
                               int offset)
    {
        Shader sh = app.getShader();

        if(sh == currentShader)
            return offset;

        int idx = offset;

        if(sh instanceof GLSLangShader)
        {
            // If the old one was GL14, just replace it entirely with the
            // new one.
            if(currentShader instanceof GLSLangShader)
            {
                GLSLangShader gl_sh = (GLSLangShader)sh;
                ShaderProgram prog = gl_sh.getShaderProgram();

                if(prog != null)
                {
                    if(currentShaderProgram != prog)
                    {
                        currentShaderProgram = prog;

                        instr.renderList[idx].renderable = currentShaderProgram;
                        instr.renderOps[idx] = RenderOp.START_SHADER_PROGRAM;
                        idx++;
                    }

                    ShaderArguments args = gl_sh.getShaderArguments();
                    if(currentShaderArgs != args)
                    {
                        currentShaderArgs = args;

                        instr.renderList[idx].renderable = currentShaderArgs;
                        instr.renderOps[idx] = RenderOp.SET_SHADER_ARGS;
                        idx++;
                    }
                }
                else
                {
                    // it's null, so effectively just turn off the old
                    // shader and set the local state here to be as though
                    // no shader is currently set.
                    instr.renderList[idx].renderable = currentShaderProgram;
                    instr.renderOps[idx] = RenderOp.STOP_SHADER_PROGRAM;
                    idx++;

                    currentShader = null;
                    currentShaderProgram = null;
                    currentShaderArgs = null;
                }
            }
            else if(currentShader instanceof GL14Shader)
            {
                instr.renderList[idx].renderable = currentShader;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                // Now push the new values, but only bother if there is
                // a program to actually run. If not, just ignore it and
                // move on as though there were no state set.
                GLSLangShader gl_sh = (GLSLangShader)sh;
                currentShaderProgram = gl_sh.getShaderProgram();

                if(currentShaderProgram != null)
                {
                    instr.renderList[idx].renderable = currentShaderProgram;
                    instr.renderOps[idx] = RenderOp.START_SHADER_PROGRAM;
                    idx++;

                    currentShaderArgs = gl_sh.getShaderArguments();

                    if(currentShaderArgs != null)
                    {
                        instr.renderList[idx].renderable = currentShaderArgs;
                        instr.renderOps[idx] = RenderOp.SET_SHADER_ARGS;
                        idx++;
                    }

                    currentShader = sh;
                }
            }
            else
            {
                GLSLangShader gl_sh = (GLSLangShader)sh;
                currentShaderProgram = gl_sh.getShaderProgram();

                // Only set a new state if we have something valid to
                // render.
                if(currentShaderProgram != null)
                {
                    instr.renderList[idx].renderable = currentShaderProgram;
                    instr.renderOps[idx] = RenderOp.START_SHADER_PROGRAM;
                    idx++;

                    currentShaderArgs = gl_sh.getShaderArguments();

                    instr.renderList[idx].renderable = currentShaderArgs;
                    instr.renderOps[idx] = RenderOp.SET_SHADER_ARGS;
                    idx++;

                    currentShader = sh;
                }
            }
        }
        else if(sh instanceof GL14Shader)
        {
            if(sh != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!sh.equals(currentShader))
                {
                    // Clear the old state if it is set
                    if(currentShader != null)
                    {
                        if(currentShader instanceof GLSLangShader)
                        {
                            instr.renderList[idx].renderable = currentShader;
                            instr.renderOps[idx] = RenderOp.STOP_SHADER_PROGRAM;
                            idx++;

                            currentShaderProgram = null;
                            currentShaderArgs = null;
                        }
                        else
                        {
                            instr.renderList[idx].renderable = currentShader;
                            instr.renderOps[idx] = RenderOp.STOP_STATE;
                            idx++;
                        }
                    }

                    instr.renderList[idx].renderable = sh;
                    instr.renderOps[idx] = RenderOp.STOP_STATE;
                    idx++;
                }

                currentShader = sh;
            }
            else
            {
                // The new shader is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentShader;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentShader = null;
            }
        }
        else if(sh != null)
        {
            // A custom shader implementation. Maybe it's Cg or something
            // customised from the extensions area such as h-anim or volume
            // rendering?
            if(!sh.equals(currentShader))
            {
                // Clear the old state if it is set
                if(currentShader != null)
                {
                    if(currentShader instanceof GLSLangShader)
                    {
                        instr.renderList[idx].renderable = currentShader;
                        instr.renderOps[idx] = RenderOp.STOP_SHADER_PROGRAM;
                        idx++;

                        currentShaderProgram = null;
                        currentShaderArgs = null;
                    }
                    else
                    {
                        instr.renderList[idx].renderable = currentShader;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }
                }

                instr.renderList[idx].renderable = sh;
                instr.renderOps[idx] = RenderOp.START_STATE;
                idx++;

                currentShader = sh;
            }
        }
        else
        {
            // incoming shader is null, so clean up anything left
            if(currentShader instanceof GLSLangShader)
            {
                GLSLangShader gl_sh = (GLSLangShader)sh;
                instr.renderList[idx].renderable = currentShaderProgram;
                instr.renderOps[idx] = RenderOp.STOP_SHADER_PROGRAM;
                idx++;

                currentShaderProgram = null;
                currentShaderArgs = null;
                currentShader = null;
            }
            else if(currentShader instanceof GL14Shader)
            {
                instr.renderList[idx].renderable = currentShader;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentShader = null;
            }
            else if (currentShader != null)
            {
                instr.renderList[idx].renderable = currentShader;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;
            }
        }

        return idx;
    }

    /**
     * Check on the material setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int updateMaterial(Appearance app,
                                 GraphicsInstructions instr,
                                 int offset)
    {
        int idx = offset;

        Material mat = app.getMaterial();
        if(mat != currentMaterial)
        {
            if(mat != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!mat.equals(currentMaterial))
                {
                    // Clear the old state if it is set
                    if(currentMaterial != null)
                    {
                        instr.renderList[idx].renderable = currentMaterial;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = mat;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentMaterial = mat;
            }
            else
            {
                // The new material is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentMaterial;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentMaterial = null;
            }
        }

        return idx;
    }

    /**
     * Check on the polygon attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int updatePolyAttribs(Appearance app,
                                    GraphicsInstructions instr,
                                    int offset)
    {
        int idx = offset;

        PolygonAttributes attr = app.getPolygonAttributes();
        if(attr != currentPolyAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentPolyAttr))
                {
                    // Clear the old state if it is set
                    if(currentPolyAttr != null)
                    {
                        instr.renderList[idx].renderable = currentPolyAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentPolyAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentPolyAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentPolyAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the line attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int updateLineAttribs(Appearance app,
                                    GraphicsInstructions instr,
                                    int offset)
    {
        int idx = offset;

        LineAttributes attr = app.getLineAttributes();
        if(attr != currentLineAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentLineAttr))
                {
                    // Clear the old state if it is set
                    if(currentLineAttr != null)
                    {
                        instr.renderList[idx].renderable = currentLineAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentLineAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentLineAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentLineAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the depth attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int updateDepthAttribs(Appearance app,
                                     GraphicsInstructions instr,
                                     int offset)
    {
        int idx = offset;

        DepthAttributes attr = app.getDepthAttributes();
        if(attr != currentDepthAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentDepthAttr))
                {
                    // Clear the old state if it is set
                    if(currentDepthAttr != null)
                    {
                        instr.renderList[idx].renderable = currentDepthAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentDepthAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentDepthAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentDepthAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the stencil attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int updateStencilAttribs(Appearance app,
                                       GraphicsInstructions instr,
                                       int offset)
    {
        int idx = offset;

        StencilAttributes attr = app.getStencilAttributes();
        if(attr != currentStencilAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentStencilAttr))
                {
                    // Clear the old state if it is set
                    if(currentStencilAttr != null)
                    {
                        instr.renderList[idx].renderable = currentStencilAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentStencilAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentStencilAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentStencilAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the polygon attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int updatePointAttribs(Appearance app,
                                     GraphicsInstructions instr,
                                     int offset)
    {
        int idx = offset;

        PointAttributes attr = app.getPointAttributes();
        if(attr != currentPointAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentPointAttr))
                {
                    // Clear the old state if it is set
                    if(currentPointAttr != null)
                    {
                        instr.renderList[idx].renderable = currentPointAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentPointAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentPointAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentPointAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the blending attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int updateBlendAttribs(Appearance app,
                                     GraphicsInstructions instr,
                                     int offset)
    {
        int idx = offset;

        BlendAttributes attr = app.getBlendAttributes();
        if(attr != currentBlendAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentBlendAttr))
                {
                    // Clear the old state if it is set
                    if(currentBlendAttr != null)
                    {
                        instr.renderList[idx].renderable = currentBlendAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentBlendAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentBlendAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentBlendAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the texture listing setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int updateTextures(Appearance app,
                                 GraphicsInstructions instr,
                                 int offset)
    {
        int idx = offset;

        int num_tex = app.numTextureUnits();

        if(textureTmp.length < num_tex)
            textureTmp = new TextureUnit[num_tex];

        app.getTextureUnits(textureTmp);

        if(num_tex != numTextures)
        {
            if(num_tex == 0)
            {
                // Clear the old textures that were set
                for(int i = 0; i < numTextures; i++)
                {
                    instr.renderList[idx].renderable = currentTextures[i];
                    instr.renderList[idx].instructions = TEX_IDS[i];
                    instr.renderOps[idx] = RenderOp.STOP_TEXTURE;
                    idx++;

                    currentTextures[i] = null;
                    textureTmp[i] = null;
                }
            }
            else if(numTextures == 0)
            {
                // There were no old textures, so just list the new ones
                for(int i = 0; i < num_tex; i++)
                {
                    instr.renderList[idx].renderable = textureTmp[i];
                    instr.renderList[idx].instructions = TEX_IDS[i];
                    instr.renderOps[idx] = RenderOp.START_TEXTURE;
                    idx++;

                    currentTextures[i] = textureTmp[i];
                    textureTmp[i] = null;
                }
            }
            else
            {
                int smaller = (numTextures < num_tex) ? numTextures : num_tex;

                for(int i = 0; i < smaller; i++)
                {
                    if(!textureTmp[i].equals(currentTextures[i]))
                    {
                        instr.renderList[idx].renderable = currentTextures[i];
                        instr.renderList[idx].instructions = TEX_IDS[i];
                        instr.renderOps[idx] = RenderOp.STOP_TEXTURE;
                        idx++;

                        instr.renderList[idx].renderable = textureTmp[i];
                        instr.renderList[idx].instructions = TEX_IDS[i];
                        instr.renderOps[idx] = RenderOp.START_TEXTURE;
                        idx++;

                        currentTextures[i] = textureTmp[i];
                        textureTmp[i] = null;
                    }
                }

                // Clean up what is left or add the new stuff
                if(numTextures < num_tex)
                {
                    for(int i = smaller; i < num_tex; i++)
                    {
                        instr.renderList[idx].renderable = textureTmp[i];
                        instr.renderList[idx].instructions = TEX_IDS[i];
                        instr.renderOps[idx] = RenderOp.START_TEXTURE;
                        idx++;

                        currentTextures[i] = textureTmp[i];
                        textureTmp[i] = null;
                    }
                }
                else
                {
                    for(int i = smaller; i < numTextures; i++)
                    {
                        instr.renderList[idx].renderable = currentTextures[i];
                        instr.renderList[idx].instructions = TEX_IDS[i];
                        instr.renderOps[idx] = RenderOp.STOP_TEXTURE;
                        idx++;

                        currentTextures[i] = null;
                        textureTmp[i] = null;
                    }
                }
            }

            numTextures = num_tex;
        }
        else if(num_tex != 0)
        {
            // same number of texture units.
            // Special case the most common situation.
            if(num_tex == 1)
            {
                if(!textureTmp[0].equals(currentTextures[0]))
                {
                    instr.renderList[idx].renderable = currentTextures[0];
                    instr.renderList[idx].instructions = TEX_IDS[0];
                    instr.renderOps[idx] = RenderOp.STOP_TEXTURE;
                    idx++;

                    instr.renderList[idx].renderable = textureTmp[0];
                    instr.renderList[idx].instructions = TEX_IDS[0];
                    instr.renderOps[idx] = RenderOp.START_TEXTURE;
                    idx++;

                    currentTextures[0] = textureTmp[0];
                    textureTmp[0] = null;
                }
            }
            else
            {
                for(int i = 0; i < numTextures; i++)
                {
                    if(!textureTmp[i].equals(currentTextures[i]))
                    {
                        instr.renderList[idx].renderable = currentTextures[i];
                        instr.renderList[idx].instructions = TEX_IDS[i];
                        instr.renderOps[idx] = RenderOp.STOP_TEXTURE;
                        idx++;

                        instr.renderList[idx].renderable = textureTmp[i];
                        instr.renderList[idx].instructions = TEX_IDS[i];
                        instr.renderOps[idx] = RenderOp.START_TEXTURE;
                        idx++;

                        currentTextures[i] = textureTmp[i];
                        textureTmp[i] = null;
                    }
                }
            }
        }

        return idx;
    }

    /**
     * At the end of the scene, we need to make sure that we've turned off all
     * the lights, clip planes and current rendering state that is left. Do
     * that now.
     *
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int cleanupObjects(GraphicsInstructions instr, int offset)
    {
        int idx = offset;

        int num_lights = currentLights.size();
        if(num_lights != 0)
        {
            if(num_lights > lightTmp.length)
                lightTmp = new Object[num_lights];

            currentLights.toArray(lightTmp);
            currentLights.clear();

            for(int i = 0; i < num_lights; i++)
            {
                VisualDetails ld = (VisualDetails)lightTmp[i];
                Integer l_id = (Integer)lightIdMap.remove(ld);

                instr.renderList[idx].renderable = ld.getRenderable();
                instr.renderList[idx].id = l_id.intValue();

                // Don't need the transform for the light stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_LIGHT;
                idx++;

                lightTmp[i] = null;
            }
        }

        // Now the clip planes.
        int num_clips = currentClipPlanes.size();
        if(num_clips != 0)
        {
            if(num_clips > clipTmp.length)
                clipTmp = new Object[num_clips];

            currentClipPlanes.toArray(clipTmp);
            currentClipPlanes.clear();

            for(int i = 0; i < num_clips; i++)
            {
                VisualDetails ld = (VisualDetails)clipTmp[i];
                Integer l_id = (Integer)clipIdMap.remove(ld);

                instr.renderList[idx].renderable = ld.getRenderable();
                instr.renderList[idx].id = l_id.intValue();

                // Don't need the transform for the clip stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_CLIP_PLANE;
                idx++;

                clipTmp[i] = null;
            }
        }

        return cleanupVisuals(instr, idx);
    }

    /**
     * Convenience method to clear all the current visual information.
     * Visuals are defined as material, shaders various attributes etc.
     *
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    protected int cleanupVisuals(GraphicsInstructions instr, int offset)
    {
        int idx = offset;

        if(currentMaterial != null)
        {
            instr.renderList[idx].renderable = currentMaterial;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentMaterial = null;
        }

        if(currentShader != null)
        {
            instr.renderList[idx].renderable = currentShader;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentShader = null;
        }

        if(currentVtxShader != null)
        {
            instr.renderList[idx].renderable = currentVtxShader;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentVtxShader = null;
        }

        if(currentFragShader != null)
        {
            instr.renderList[idx].renderable = currentFragShader;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentFragShader = null;
        }

        if(currentPolyAttr != null)
        {
            instr.renderList[idx].renderable = currentPolyAttr;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentPolyAttr = null;
        }

        if(currentLineAttr != null)
        {
            instr.renderList[idx].renderable = currentLineAttr;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentLineAttr = null;
        }

        if(currentPointAttr != null)
        {
            instr.renderList[idx].renderable = currentPointAttr;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentPointAttr = null;
        }

        if(currentBlendAttr != null)
        {
            instr.renderList[idx].renderable = currentBlendAttr;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentBlendAttr = null;
        }

        for(int i = 0; i < numTextures; i++)
        {
            instr.renderList[idx].renderable = currentTextures[i];
            instr.renderList[idx].instructions = TEX_IDS[i];
            instr.renderOps[idx] = RenderOp.STOP_TEXTURE;
            idx++;

            currentTextures[i] = null;
        }

        if(currentFog != null)
        {
            instr.renderList[idx].renderable = currentFog;
            instr.renderOps[idx] = RenderOp.STOP_FOG;
            idx++;

            currentFog = null;
        }

        numTextures = 0;
        currentAppearance = null;

        return idx;
    }
}
