/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004 - 2006
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline.graphics;

// External imports
import javax.vecmath.Matrix4f;

// Local imports
import org.j3d.aviatrix3d.*;

/**
 * Handles the scenegraph maintenance and culling operations.
 * <p>
 *
 * The culling phase generates a list of nodes to render.
 * A future optimization will sort the render list by OGL state.
 * <p>
 *
 * By default, the implementation will walk into shape3Ds looking for any
 * offscreen textures to be rendered. If you know that you do not have any
 * in the scene, then you can set an internal flag to not look for them, thus
 * achieving a simple performance boost.
 * <p>
 *
 * Since this is a null culling stage, and thus no culling is performed, if the
 * scene graph contains instances of Cullable, they are ignored and traversal
 * stops at that point.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 3.5 $
 */
public class NullCullStage extends BaseCullStage
{
    /**
     * Create a basic instance of this class with the list assuming there are
     * no off-screen buffers in use for the initial internal setup.
     */
    public NullCullStage()
    {
        this(LIST_START_LENGTH);
    }

    /**
     * Create a basic instance of this class with the list initial internal
     * setup for the given number of renderable surfaces. The size is just an
     * initial esstimate, and is used for optimisation purposes to prevent
     * frequent array reallocations internally. As such, the number does not
     * have to be perfect, just good enough.
     *
     * @param numSurfaces Total number of surfaces to prepare rendering for
     */
    public NullCullStage(int numSurfaces)
    {
        super(numSurfaces);
    }

    //---------------------------------------------------------------
    // Methods defined by BaseCullStage
    //---------------------------------------------------------------

    /**
     * Update and cull the scenegraph. This generates an ordered list
     * of nodes to render. It will not return until the culling is complete.
     *
     * @param scene The scene instance to cull
     * @param viewIndex The index of the viewport in the parent
     *    ViewportCollection
     * @param layerIndex The index of the layer within the viewport
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    protected void cullScene(SimpleScene scene,
                             int subsceneId,
                             int layerId,
                             int viewIndex,
                             int layerIndex,
                             int buffer)
    {
        // Buffer ID is not used currently.
        lastLight = 0;
        lastClip = 0;
        lastTxStack = 0;
        lastFogStack = 0;
        activeParent = null;

        SceneRenderBucket bucket =
            outputLayers[subsceneId][layerId].layers[viewIndex].scenes[layerIndex];

        workCullList = bucket.nodes;
        validSceneParents[0][0] = null;
        validSceneParents[0][1] = null;

        Node node = scene.getRenderedGeometry();
        fillEnvData(scene, bucket.data, buffer);
        bucket.data.layerId = layerId;
        bucket.data.subLayerId = layerIndex;

        if(terminate)
            return;

        if(node instanceof Group)
        {
            // set the top matrix stack
            if(node instanceof TransformGroup)
                ((TransformGroup)node).getTransform(transformStack[0]);
            else
                transformStack[0].setIdentity();

            bucket.numNodes = findAllNodes((Group)node, 0, buffer);
            bucket.nodes = workCullList;
            lastOutputList++;
        }
        else if(node instanceof SharedNode)
        {
            // set the top matrix stack back to identity
            transformStack[0].setIdentity();
            bucket.numNodes = findNextNode((SharedNode)node,  0, buffer);
            bucket.nodes = workCullList;
            lastOutputList++;
        }
        else if((node instanceof Shape3D) && ((Shape3D)node).isVisible())
        {
            // Walk into the shape and check that we don't have any
            // offscreen textures to render.
            if(checkOffscreens)
                checkForOffscreens((Shape3D)node, buffer);

            workCullList[0].renderable = (Shape3D)node;
            workCullList[0].transform.setIdentity();
            workCullList[0].numLights = 0;
            workCullList[0].numClipPlanes = 0;

            bucket.numNodes = 1;
            bucket.nodes = workCullList;
            lastOutputList++;
        }
    }

    /**
     * Update and cull the a single pass from a multipass rendering. This
     * generates an ordered list of nodes to render in the same was as a normal
     * scene, but with fewer items updated, such as only a single background
     * for all passes. It will not return until the culling is complete.
     *
     * @param pass The rendering pass instance to cull
     * @param subsceneId The index of the subscene in the viewport output
     * @param layerId The index of the layer in the output
     * @param viewIndex The index of the viewport in the parent
     *    ViewportCollection
     * @param layerIndex The index of the layer within the viewport
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    protected void cullRenderPass(RenderPass pass,
                                  int passNumber,
                                  int subsceneId,
                                  int layerId,
                                  int viewIndex,
                                  int layerIndex,
                                  int buffer)
    {
        lastLight = 0;
        lastClip = 0;
        lastTxStack = 0;
        lastFogStack = 0;
        activeParent = null;

        // Assumes main scene only. Doesn't yet handle nested texture rendering
        ViewportLayerCollection c =
            outputLayers[subsceneId][layerId].layers[viewIndex];

        MultipassDetails bucket =
            c.multipass[c.numMultipass].mainScene;

        workCullList = bucket.nodes[passNumber];
        validSceneParents[0][0] = null;
        validSceneParents[0][1] = null;

        Node node = pass.getRenderedGeometry();
        fillEnvData(pass,
                    bucket.data[passNumber],
                    bucket.buffers[passNumber],
                    buffer);

        bucket.data[passNumber].layerId = layerId;
        bucket.data[passNumber].subLayerId = layerIndex;

        if(terminate)
            return;

        if(node instanceof Group)
        {
            // set the top matrix stack back to identity
            if(node instanceof TransformGroup)
                ((TransformGroup)node).getTransform(transformStack[0]);
            else
                transformStack[0].setIdentity();

            bucket.numNodes[passNumber] = findAllNodes((Group)node, 0, buffer);
            bucket.nodes[passNumber] = workCullList;

            lastOutputList++;
        }
        else if(node instanceof SharedNode)
        {
            // set the top matrix stack back to identity
            transformStack[0].setIdentity();

            bucket.numNodes[passNumber] = findNextNode((SharedNode)node,  0, buffer);
            bucket.nodes[passNumber] = workCullList;
            lastOutputList++;
        }
        else if((node instanceof Shape3D) && ((Shape3D)node).isVisible())
        {
            // Walk into the shape and check that we don't have any
            // offscreen textures to render.
            if(checkOffscreens)
                checkForOffscreens((Shape3D)node, buffer);

            workCullList[0].renderable = (Shape3D)node;
            workCullList[0].transform.setIdentity();
            workCullList[0].numLights = 0;
            workCullList[0].numClipPlanes = 0;

            bucket.numNodes[passNumber] = 1;
            bucket.nodes[passNumber] = workCullList;
            lastOutputList++;
        }
    }

    /**
     * Update and cull the scenegraph of a single node without doing any other
     * processing. This generates an ordered list
     * of nodes to render. It will not return until the culling is complete.
     *
     * @param node The root node to start the cull from
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     * @return The index of the last item on the cull list
     */
    protected int cullSingle(Group node, int buffer)
    {
        // set the top matrix stack back to identity
        transformStack[0].setIdentity();

        if(node instanceof TransformGroup)
            ((TransformGroup)node).getTransform(transformStack[0]);

        return findAllNodes((Group)node, 0, buffer);
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Recursive walk of the tree to find all the renderable nodes.
     *
     * @param group The parent node to walk into
     * @param cullEndIndex The current last item on the cull list
     * @param buffer The id of the buffer to process for this round
     * @return The index of the last item on the cull list
     */
    private int findAllNodes(Group group, int cullEndIndex, int buffer)
    {
        if(terminate)
            return 0;

        Node[] kids  = group.getRenderableChild();
        int size = group.numRenderableChildren();

        int num_lights = 0;
        int num_clips = 0;
        int ret_val = cullEndIndex;
        boolean have_local_fog = false;

        // Find all the lights and clip planes at this level first. Place
        // them first into the queue
        for(int i = 0; i < size; i++)
        {
            if(kids[i] instanceof Light)
            {
                Light l = (Light)kids[i];
                if(!l.isEnabled())
                    continue;

                resizeLightList();
                lightList[lastLight] = l;

                Matrix4f mat = transformStack[lastTxStack];

                // Transpose the matrix in place as it is being copied
                lightTxList[lastLight][0] = (float)mat.m00;
                lightTxList[lastLight][1] = (float)mat.m10;
                lightTxList[lastLight][2] = (float)mat.m20;
                lightTxList[lastLight][3] = (float)mat.m30;

                lightTxList[lastLight][4] = (float)mat.m01;
                lightTxList[lastLight][5] = (float)mat.m11;
                lightTxList[lastLight][6] = (float)mat.m21;
                lightTxList[lastLight][7] = (float)mat.m31;

                lightTxList[lastLight][8] = (float)mat.m02;
                lightTxList[lastLight][9] = (float)mat.m12;
                lightTxList[lastLight][10] = (float)mat.m22;
                lightTxList[lastLight][11] = (float)mat.m32;

                lightTxList[lastLight][12] = (float)mat.m03;
                lightTxList[lastLight][13] = (float)mat.m13;
                lightTxList[lastLight][14] = (float)mat.m23;
                lightTxList[lastLight][15] = (float)mat.m33;

                lastLight++;
                num_lights++;
            }
            else if(kids[i] instanceof ClipPlane)
            {
                ClipPlane cp = (ClipPlane)kids[i];
                if(!cp.isEnabled())
                    continue;

                resizeClipList();
                clipList[lastClip] = cp;

                Matrix4f mat = transformStack[lastTxStack];

                // Transpose the matrix in place as it is being copied
                clipTxList[lastClip][0] = (float)mat.m00;
                clipTxList[lastClip][1] = (float)mat.m10;
                clipTxList[lastClip][2] = (float)mat.m20;
                clipTxList[lastClip][3] = (float)mat.m30;

                clipTxList[lastClip][4] = (float)mat.m01;
                clipTxList[lastClip][5] = (float)mat.m11;
                clipTxList[lastClip][6] = (float)mat.m21;
                clipTxList[lastClip][7] = (float)mat.m31;

                clipTxList[lastClip][8] = (float)mat.m02;
                clipTxList[lastClip][9] = (float)mat.m12;
                clipTxList[lastClip][10] = (float)mat.m22;
                clipTxList[lastClip][11] = (float)mat.m32;

                clipTxList[lastClip][12] = (float)mat.m03;
                clipTxList[lastClip][13] = (float)mat.m13;
                clipTxList[lastClip][14] = (float)mat.m23;
                clipTxList[lastClip][15] = (float)mat.m33;

                lastClip++;
                num_clips++;
            }
            else if(kids[i] instanceof Fog)
            {
                Fog fog = (Fog)kids[i];
                if(fog.isEnabled() && !fog.isGlobalOnly() && !have_local_fog)
                {
                    resizeFogStack();

                    have_local_fog = true;
                    lastFogStack++;
                    fogStack[lastFogStack] = fog;
                }
            }
        }

        for(int i = 0; i < size && !terminate; i++)
        {
            if(kids[i] instanceof Group)
            {
                // If a tg, push the new TX onto the stack
                boolean is_tx = (kids[i] instanceof TransformGroup);

                if(is_tx)
                {
                    resizeStack();
                    TransformGroup tg = (TransformGroup)kids[i];
                    tg.getTransform(transformStack[lastTxStack + 1]);

                    transformStack[lastTxStack + 1].mul(transformStack[lastTxStack],
                                                      transformStack[lastTxStack + 1]);
                    lastTxStack++;
                }

                ret_val = findAllNodes((Group)kids[i], ret_val, buffer);

                // Now pop the stacks

                if(is_tx)
                    lastTxStack--;
            }
            else if(kids[i] instanceof Shape3D)
            {
                // Check the visibility state and ignore if not visible.
                if(!((Shape3D)kids[i]).isVisible())
                    continue;

                // Walk into the shape and check that we don't have any
                // offscreen textures to render.
                if(checkOffscreens)
                    checkForOffscreens((Shape3D)kids[i], buffer);

                resizeCullList(ret_val);
                workCullList[ret_val].renderable = (Shape3D)kids[i];
                workCullList[ret_val].localFog = fogStack[lastFogStack];

                Matrix4f mat = transformStack[lastTxStack];

                // Transpose the matrix in place as it is being copied
                workCullList[ret_val].transform.set(mat);
                int src_size = (workCullList[ret_val].lights == null) ?
                                0 :
                                workCullList[ret_val].lights.length;

                if(src_size < lastLight)
                {
                    // up the size of the array
                    VisualDetails[] tmp = new VisualDetails[lastLight];
                    if(src_size != 0)
                    {
                        System.arraycopy(workCullList[ret_val].lights,
                                         0,
                                         tmp,
                                         0,
                                         src_size);
                    }

                    for(int j = src_size; j < lastLight; j++)
                        tmp[j] = new VisualDetails();

                    workCullList[ret_val].lights = tmp;
                }

                // copy in the light information.
                VisualDetails[] l_tmp = workCullList[ret_val].lights;
                workCullList[ret_val].numLights = lastLight;

                for(int j = 0; j < lastLight; j++)
                    l_tmp[j].update(lightList[j], lightTxList[j]);

                // Same thing again but for clip planes
                src_size = (workCullList[ret_val].clipPlanes == null) ?
                            0 :
                            workCullList[ret_val].clipPlanes.length;

                if(src_size < lastClip)
                {
                    // up the size of the array
                    VisualDetails[] tmp = new VisualDetails[lastClip];
                    if(src_size != 0)
                    {
                        System.arraycopy(workCullList[ret_val].clipPlanes,
                                         0,
                                         tmp,
                                         0,
                                         src_size);
                    }

                    for(int j = src_size; j < lastClip; j++)
                        tmp[j] = new VisualDetails();

                    workCullList[ret_val].clipPlanes = tmp;
                }

                // copy in the light information.
                VisualDetails[] c_tmp = workCullList[ret_val].clipPlanes;
                workCullList[ret_val].numClipPlanes = lastClip;

                for(int j = 0; j < lastClip; j++)
                    c_tmp[j].update(clipList[j], clipTxList[j]);

                ret_val++;
            }
            else if(kids[i] instanceof SharedNode)
            {
                ret_val = findNextNode((SharedNode)kids[i],  ret_val, buffer);
            }
        }

        // Pop the valid lights from the stack
        lastLight -= num_lights;
        lastClip -= num_clips;

        if(have_local_fog)
        {
            fogStack[lastFogStack] = null;
            lastFogStack--;
        }

        return ret_val;
    }

    /**
     * From the given SharedNode instance keep finding a non-sharedNode
     * instance, before continuing on the recursion
     *
     * @param root The parent node to walk into
     * @param cullEndIndex The current last item on the cull list
     * @param buffer The id of the buffer to process for this round
     * @return The index of the last item on the cull list
     */
    private int findNextNode(SharedNode root, int cullEndIndex, int buffer)
    {
        if(terminate)
            return 0;

        int ret_val = cullEndIndex;

        Node current = root.getChild();

        while(current != null && current instanceof SharedNode)
            current = ((SharedNode)current).getChild();

        // Now just continue on normally. If the result of the above was a
        // null then it just falls off the end of this if/else naturally.
        if(current instanceof Group)
        {
            // If a tg, push the new TX onto the stack
            boolean is_tx = (current instanceof TransformGroup);

            if(is_tx)
            {
                resizeStack();
                TransformGroup tg = (TransformGroup)current;
                tg.getTransform(transformStack[lastTxStack + 1]);

                transformStack[lastTxStack + 1].mul(transformStack[lastTxStack],
                                                  transformStack[lastTxStack + 1]);
                lastTxStack++;
            }

            ret_val = findAllNodes((Group)current, cullEndIndex, buffer);

            // Now pop the stack.
            if(is_tx)
                lastTxStack--;
        }
        else if(current instanceof Shape3D)
        {
            // Check the visibility state and ignore if not visible.
            if(!((Shape3D)current).isVisible())
                return ret_val;

            // Walk into the shape and check that we don't have any
            // offscreen textures to render.
            if(checkOffscreens)
                checkForOffscreens((Shape3D)current, buffer);

            resizeCullList(cullEndIndex);
            workCullList[cullEndIndex].renderable = (Shape3D)current;
            workCullList[cullEndIndex].localFog = fogStack[lastFogStack];

            Matrix4f mat = transformStack[lastTxStack];

            // Transpose the matrix in place as it is being copied
            workCullList[cullEndIndex].transform.set(mat);

            int src_size = (workCullList[cullEndIndex].lights == null) ?
                            0 :
                            workCullList[cullEndIndex].lights.length;

            if(src_size < lastLight)
            {
                // up the size of the array

                VisualDetails[] tmp = new VisualDetails[lastLight];
                if(src_size != 0)
                {
                    System.arraycopy(workCullList[cullEndIndex].lights,
                                     0,
                                     tmp,
                                     0,
                                     src_size);
                }

                for(int j = src_size; j < lastLight; j++)
                    tmp[j] = new VisualDetails();

                workCullList[cullEndIndex].lights = tmp;
            }

            // copy in the light information.
            VisualDetails[] l_tmp = workCullList[cullEndIndex].lights;
            workCullList[cullEndIndex].numLights = lastLight;

            for(int j = 0; j < lastLight; j++)
                l_tmp[j].update(lightList[j], lightTxList[j]);

            // Same again, but with clip information
            src_size = (workCullList[cullEndIndex].clipPlanes == null) ?
                       0 :
                       workCullList[cullEndIndex].clipPlanes.length;

            if(src_size < lastClip)
            {
                // up the size of the array

                VisualDetails[] tmp = new VisualDetails[lastClip];
                if(src_size != 0)
                {
                    System.arraycopy(workCullList[cullEndIndex].clipPlanes,
                                     0,
                                     tmp,
                                     0,
                                     src_size);
                }

                for(int j = src_size; j < lastClip; j++)
                    tmp[j] = new VisualDetails();

                workCullList[cullEndIndex].clipPlanes = tmp;
            }

            // copy in the clip information.
            VisualDetails[] c_tmp = workCullList[cullEndIndex].clipPlanes;
            workCullList[cullEndIndex].numClipPlanes = lastClip;

            for(int j = 0; j < lastClip; j++)
                c_tmp[j].update(clipList[j], clipTxList[j]);

            cullEndIndex++;

            ret_val++;
        }

        return ret_val;
    }
}
