/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.renderer.aviatrix3d.nodes;

// External imports
import javax.vecmath.Matrix4f;
import javax.vecmath.Vector4f;
import javax.vecmath.Vector3f;
import javax.vecmath.Point3f;
import javax.vecmath.AxisAngle4f;

// Local imports
import org.j3d.aviatrix3d.BoundingBox;
import org.j3d.aviatrix3d.Group;
import org.j3d.aviatrix3d.Cullable;
import org.j3d.aviatrix3d.CullInstructions;

import org.j3d.aviatrix3d.picking.BoundingVolume;

/**
 * <b>Not available at this time</b>
 * Provides local billboard capabilities that orient the children object
 * towards the camera location.
 * <p>
 *
 * This implementation provides view-dependent correctness based on the
 * individual traversal path to the node from the root of the scene graph.
 * Even under a shared subgraph, or different rendering pipelines with different
 * view frustums, each object traversal path will look visually correct.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 1.2 $
 */
public class Billboard extends Group implements Cullable
{
    /** Vector pointing upwards */
    private static final Vector3f Y_UP = new Vector3f(0, 1, 0);

    /** Vector pointing upwards */
    private static final Vector3f Z_UP = new Vector3f(0, 0, 1);

    /** Position representing the local origin */
    private static final Point3f ORIGIN = new Point3f(0, 0, 0);

    /** The X component of the axis of rotation */
    private float xAoR;

    /** The X component of the axis of rotation */
    private float yAoR;

    /** The X component of the axis of rotation */
    private float zAoR;

    /**
     * Construct a new billboard with the axis of rotation set to 0,0,0.
     */
    public Billboard()
    {
        xAoR = 0;
        yAoR = 0;
        zAoR = 0;
    }

    //----------------------------------------------------------
    // Methods defined by Cullable
    //----------------------------------------------------------

    /**
     * Check this node for children to traverse. The angular resolution is
     * defined as Field Of View (in radians) / viewport width in pixels.
     *
     * @param output Fill in the child information here
     * @param vworldTx The transformation from the root of the scene to
     *    this node according to the current traversal path
     * @param viewTransform The transformation from the root of the scene
     *    graph to the active viewpoint
     * @param frustumPlanes Listing of frustum planes in the order: right,
     *    left, bottom, top, far, near
     * @param angularRes Angular resolution of the screen, or 0 if not
     *    calculable from the available data.
     */
    public void cullChildren(CullInstructions output,
                             Matrix4f vworldTx,
                             Matrix4f viewTransform,
                             Vector4f[] frustumPlanes,
                             float angularRes)
    {
        if(BoundingVolume.FRUSTUM_ALLOUT ==
            bounds.checkIntersectionFrustum(frustumPlanes, vworldTx))
        {
            output.hasTransform = false;
            output.instructions = null;
            output.numChildren = 0;
            return;
        }

        // All positions are transformed to the world coordinate
        // system to avoid matrix inversions

        // Compute camera - object vector
        double camera_x = viewTransform.m03 - vworldTx.m03;
        double camera_y = viewTransform.m13 - vworldTx.m13;
        double camera_z = viewTransform.m23 - vworldTx.m23;
        double d = Math.sqrt(camera_x*camera_x+camera_y*camera_y+camera_z*camera_z);
        camera_x /= d;
        camera_y /= d;
        camera_z /= d;

        // Transformed up vector (0,1,0)
        double up_x = vworldTx.m01;
        double up_y = vworldTx.m11;
        double up_z = vworldTx.m21;
        d = Math.sqrt(up_x*up_x+up_y*up_y+up_z*up_z);
        up_x /= d;
        up_y /= d;
        up_z /= d;

        double siny = camera_x*up_x+camera_y*up_y+camera_z*up_z;
        double cosy = 1 - siny*siny;
        double sin = 0;
        double cos = 1;
        if (cosy < 1e-5)
        {
            cosy = 0;
            siny = 1;
        }
        else
        {
            cosy = Math.sqrt(cosy);

            // Project camera into x-z plane
            camera_x -= up_x*siny;
            camera_y -= up_y*siny;
            camera_z -= up_z*siny;

            // Transformed front vector (0,0,1)
            double front_x = vworldTx.m02;
            double front_y = vworldTx.m12;
            double front_z = vworldTx.m22;

            // Front dot camera, normalized
            cos = camera_x*front_x + camera_y*front_y + camera_z*front_z;
            cos /= cosy*Math.sqrt(front_x*front_x+front_y*front_y+front_z*front_z);

            // Transformed right vector (1,0,0)
            double right_x = vworldTx.m00;
            double right_y = vworldTx.m10;
            double right_z = vworldTx.m20;

            // Right dot camera, normalized
            sin = camera_x*right_x + camera_y*right_y + camera_z*right_z;
            sin /= cosy*Math.sqrt(right_x*right_x+right_y*right_y+right_z*right_z);
        }

        // Set matrix values
        if (point_mode)
        {
            output.localTransform.m00 =  (float)cos;
            output.localTransform.m01 = -(float)(sin*siny);
            output.localTransform.m02 =  (float)(sin*cosy);

            output.localTransform.m10 =   0;
            output.localTransform.m11 =  (float)cosy;
            output.localTransform.m12 =  (float)siny;

            output.localTransform.m20 = -(float)sin;
            output.localTransform.m21 = -(float)(cos*siny);
            output.localTransform.m22 =  (float)(cos*cosy);

            output.localTransform.m03 = output.localTransform.m13 = output.localTransform.m23 = 0;
            output.localTransform.m30 = output.localTransform.m31 = output.localTransform.m32 = 0;
            output.localTransform.m33 = 1;
        }
        else
        {
            output.localTransform.m00 = output.localTransform.m22 = (float)cos;
            output.localTransform.m02 =  (float)sin;
            output.localTransform.m20 = -(float)sin;
            output.localTransform.m11 = output.localTransform.m33 = 1;
            output.localTransform.m01 = output.localTransform.m10 = 0;
            output.localTransform.m03 = output.localTransform.m30 = 0;
            output.localTransform.m12 = output.localTransform.m21 = 0;
            output.localTransform.m13 = output.localTransform.m31 = 0;
            output.localTransform.m23 = output.localTransform.m32 = 0;
        }

        // And the rest of the output data...
        output.hasTransform = true;

        if(output.children.length < lastChild)
            output.resizeChildren(lastChild);

        for(int i=0; i < lastChild; i++)
            output.children[i] = childList[i];

        output.instructions = null;
        output.numChildren = lastChild;
    }

    //----------------------------------------------------------
    // Methods defined by Node
    //----------------------------------------------------------

    /**
     * Implementation of recompute bounds.
     * Expands the boundingbox to compensate for rotation of the children
     */
    protected void recomputeBounds()
    {
        if(!implicitBounds)
          return;

        super.recomputeBounds();

        if(bounds == INVALID_BOUNDS)
            return;

        bounds.getExtents(wkVec1, wkVec2);
        float x = (wkVec1[0] + wkVec2[0])/2;
        float y = (wkVec1[1] + wkVec2[1])/2;
        float z = (wkVec1[2] + wkVec2[2])/2;
        float dx = x - wkVec1[0];
        float dy = y - wkVec1[1];
        float dz = z - wkVec1[2];

        if(point_mode)
            dx = dy = dz = (float)Math.sqrt(dx*dx+dy*dy+dz*dz);
        else
            dx = dz = (float)Math.sqrt(dx*dx+dz*dz);

        BoundingBox bbox = (BoundingBox)bounds;
        bbox.setMinimum(x-dx, y-dy, z-dz);
        bbox.setMaximum(x+dx, y+dy, z+dz);
    }

    //----------------------------------------------------------
    // Local Methods
    //----------------------------------------------------------

    /**
     * Set the axis of rotation that this billboard operates about.
     */
}
