/*****************************************************************************
 *                     SINTEF Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.renderer.aviatrix3d.nodes;

// External imports
import javax.vecmath.*;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.aviatrix3d.picking.BoundingVolume;

/**
 * <b>Not available at this time</b>
 * A grouping node structure that controls which children are being rendered
 * based on distance from the user.
 *
 * @author Rune Aasgaard, (c) SINTEF
 * @version 1.0
 */
public class LODGroup extends BaseNode
    implements Cullable
{
    /** Message for the index provided being out of range */
    private static final String CHILD_IDX_ERR =
      "Index provided > last valid index";

    /** The child that is currently selected for rendering */
    private Node[] renderedChild;

    /** The array of child values used for LOD computation */
    private float[] valueArr;

    /**
     * Should the values be interpreted as node - eye distance or
     * size of smallest visible feature
     */
    private boolean distanceMode;

    /**
     * Create a new LOD implementation that uses distance as the determining
     * visual factor.
     */
    public LODGroup()
    {
        this(false);
    }

    public LODGroup(boolean distanceMode)
    {
        this.distanceMode = distanceMode;

        valueArr = new float[childList.length];

        for (int i=0; i< valueArr.length; ++i)
          valueArr[i] = Float.MAX_VALUE;

        renderedChild = new Node[1];
    }

    //-------------------------------------------------
    // Methods defined by Cullable
    //-------------------------------------------------

  private float [] c = new float[3];

    /**
     * Check this node for children to traverse. The angular resolution is
     * defined as Field Of View (in radians) / viewport width in pixels.
     *
     * @param output Fill in the child information here
     * @param vworldTx The transformation from the root of the scene to
     *    this node according to the current traversal path
     * @param viewTransform The transformation from the root of the scene
     *    graph to the active viewpoint
     * @param frustumPlanes Listing of frustum planes in the order: right,
     *    left, bottom, top, far, near
     * @param angularRes Angular resolution of the screen, or 0 if not
     *    calculable from the available data.
     */
    public void cullChildren(CullInstructions output,
                             Matrix4f vworldTx,
                             Matrix4f viewTransform,
                             Vector4f[] frustumPlanes,
                             float angularRes)
    {
        if(BoundingVolume.FRUSTUM_ALLOUT ==
           bounds.checkIntersectionFrustum(frustumPlanes, vworldTx))
        {
            output.hasTransform = false;
            output.instructions = null;
            output.numChildren = 0;
            return;
        }

        double camera_x = viewTransform.m03 - vworldTx.m03;
        double camera_y = viewTransform.m13 - vworldTx.m13;
        double camera_z = viewTransform.m23 - vworldTx.m23;

        if(distanceMode)
            angularRes = 1;
        else
            angularRes = (float)Math.toRadians(angularRes);

        int render_child = -1;
        for (int i = 0; i < lastChild; ++i)
        {
            BoundingVolume ch_b = childList[i].getBounds();
            if(ch_b == null || ch_b instanceof BoundingVoid)
                continue;

            ch_b.getCenter(c);
            float radius = 0;
            if (ch_b instanceof BoundingSphere)
                radius = ((BoundingSphere)ch_b).getRadius();
            else
            {
                bounds.getExtents(wkVec1, wkVec2);
                wkVec1[0] = c[0] - wkVec1[0];
                wkVec1[1] = c[1] - wkVec1[1];
                wkVec1[2] = c[2] - wkVec1[2];
                wkVec2[0] -= c[0];
                wkVec2[1] -= c[1];
                wkVec2[2] -= c[2];
                if (wkVec1[0] < wkVec2[0]) wkVec1[0] = wkVec2[0];
                if (wkVec1[1] < wkVec2[1]) wkVec1[1] = wkVec2[1];
                if (wkVec1[2] < wkVec2[2]) wkVec1[2] = wkVec2[2];
                radius = (float)Math.sqrt(wkVec1[0]*wkVec1[0]+wkVec1[1]*wkVec1[1]+wkVec1[2]*wkVec1[2]);
            }

            c[0] -= camera_x;
            c[1] -= camera_y;
            c[2] -= camera_z;

            float dist = (float)Math.sqrt(c[0]*c[0] + c[1]*c[1] + c[2]*c[2]) - radius;

            if(dist * angularRes <= valueArr[i])
            {
                render_child = i;
                break;
            }
        }

        output.hasTransform = false;
        output.instructions = null;

        if(render_child >= 0)
        {
            if (output.children == null || output.children.length < 1)
                output.resizeChildren(1);

            output.numChildren = 1;
            output.children[0] = childList[render_child];
        }
        else
            output.numChildren = 0;
    }

    //-------------------------------------------------
    // Methods defined by Group
    //-------------------------------------------------

    /**
     * Add a new child to the group.
     */
    public void addChild(Node newChild)
    {
        super.addChild(newChild);

        if(childList.length > valueArr.length)
        {
            float tmp_fa[] = new float[childList.length];
            for (int i = 0; i < lastChild; ++i)
                tmp_fa[i] = valueArr[i];

            for (int i = lastChild; i < tmp_fa.length; ++i)
                tmp_fa[i] = Float.MAX_VALUE;

            valueArr = tmp_fa;
        }
    }


    public void removeChild(int idx)
    {
        super.removeChild(idx);
        for (int i = idx; i < lastChild; ++i)
          valueArr[i] = valueArr[i + 1];

        valueArr[lastChild] = Float.MAX_VALUE;
    }

    /**
    * Get the list of children that are valid to be rendered according to
    * the rules of the grouping node.
    *
    * @return an array of nodes
    */
    public Node[] getRenderableChild()
    {
        renderedChild[0] = childList[0];
        return renderedChild;
    }

    /**
    * Returns the number of valid renderable children to process.If there are
    * no valid renderable children return -1.
    *
    * @return A number greater than or equal to zero
    */
    public int numRenderableChildren()
    {
        if(lastChild < 1)
            return 0;
        else
            return 1;
    }

    /**
    * Removes all children from the group.
    *
    * @throws InvalidWriteTimingException An attempt was made to write outside
    *   of the NodeUpdateListener callback method
    */
    public void removeAllChildren()
        throws InvalidWriteTimingException
    {
        super.removeAllChildren();

        renderedChild[0] = null;
        for(int i = lastChild; i < valueArr.length; ++i)

          valueArr[i] = Float.MAX_VALUE;
    }

    //-------------------------------------------------
    // Local methods
    //-------------------------------------------------

    /**
     * Get the range set at the given child index.
     *
     * @param idx The range index to get the value from
     * @throws IndexOutOfBoundsException The index is negative or past the
     *   last child
     */
    public float getRange(int idx)
    {
        if(idx < 0 || idx > lastChild)
            throw new IndexOutOfBoundsException(CHILD_IDX_ERR);

        return valueArr[idx];
    }

    /**
     * Set the range at the given child index. The value at this index is
     * required to be monotonically increasing relative to the previous and next
     * index.
     *
     * @param idx The range index to get the value from
     * @throws IndexOutOfBoundsException The index is negative or past the
     *   last child
     * @throws IllegalArgumentException The value is not monotonically
     *   increasing
     */
    public void setRange(int idx, float value)
    {
        if(idx < 0 || idx > lastChild)
            throw new IndexOutOfBoundsException(CHILD_IDX_ERR);

        if((idx+1 < lastChild && valueArr[idx+1] <= value) ||
            (idx-1 >= 0        && valueArr[idx-1] >= value))
            throw new IllegalArgumentException("Value not monotonously increasing");

        valueArr[idx] = value;
    }

    /**
     * Find out which of the two modes are being used by this LOD.
     */
    public boolean isDistanceMode()
    {
        return distanceMode;
    }
}
